import abc
import os
import struct
from collections.abc import Callable
from functools import cached_property
from itertools import chain, count
from typing import TYPE_CHECKING, Any

import numpy as np

from yt._typing import FieldKey
from yt.config import ytcfg
from yt.funcs import mylog
from yt.utilities.cython_fortran_utils import FortranFile

from .field_handlers import HandlerMixin
from .io import (
    _ramses_particle_binary_file_handler,
    _ramses_particle_csv_file_handler,
    _read_part_binary_file_descriptor,
    _read_part_csv_file_descriptor,
    convert_ramses_conformal_time_to_physical_time,
)

if TYPE_CHECKING:
    from yt.frontends.ramses.data_structures import RAMSESDomainSubset

PARTICLE_HANDLERS: set[type["ParticleFileHandler"]] = set()


def get_particle_handlers():
    return PARTICLE_HANDLERS


def register_particle_handler(ph):
    PARTICLE_HANDLERS.add(ph)


class ParticleFileHandler(abc.ABC, HandlerMixin):
    """
    Abstract class to handle particles in RAMSES. Each instance
    represents a single file (one domain).

    To add support to a new particle file, inherit from this class and
    implement all functions containing a `NotImplementedError`.

    See `SinkParticleFileHandler` for an example implementation."""

    _file_type = "particle"

    ## These properties are static
    # The name to give to the particle type
    ptype: str

    # The name of the file(s).
    fname: str

    # The name of the file descriptor (if any)
    file_descriptor: str | None = None

    # The attributes of the header
    attrs: tuple[tuple[str, int, str], ...]

    # A list of tuple containing the field name and its type
    known_fields: list[FieldKey]

    # The function to employ to read the file
    reader: Callable[
        ["ParticleFileHandler", "RAMSESDomainSubset", list[tuple[str, str]], int],
        dict[tuple[str, str], np.ndarray],
    ]

    # Name of the config section (if any)
    config_field: str | None = None

    ## These properties are computed dynamically
    # Mapping from field to offset in file
    _field_offsets: dict[tuple[str, str], int]

    # Mapping from field to the type of the data (float, integer, ...)
    _field_types: dict[tuple[str, str], str]

    # Number of particle in the domain
    _local_particle_count: int

    # The header of the file
    _header: dict[str, Any]

    def __init_subclass__(cls, *args, **kwargs):
        """
        Registers subclasses at creation.
        """
        super().__init_subclass__(*args, **kwargs)

        if cls.ptype is not None:
            register_particle_handler(cls)

        cls._unique_registry = {}
        return cls

    def __init__(self, domain):
        self.setup_handler(domain)

        # Attempt to read the list of fields from the config file
        if self.config_field and ytcfg.has_section(self.config_field):
            cfg = ytcfg.get(self.config_field, "fields")
            known_fields = []
            for c in (_.strip() for _ in cfg.split("\n") if _.strip() != ""):
                field, field_type = (_.strip() for _ in c.split(","))
                known_fields.append((field, field_type))
            self.known_fields = known_fields

    @abc.abstractmethod
    def read_header(self):
        """
        This function is called once per file. It should:

        * read the header of the file and store any relevant information
        * detect the fields in the file
        * compute the offsets (location in the file) of each field

        It is in charge of setting `self.field_offsets` and `self.field_types`.

        * `_field_offsets`: dictionary: tuple -> integer
            A dictionary that maps `(type, field_name)` to their
            location in the file (integer)
        * `_field_types`: dictionary: tuple -> character
            A dictionary that maps `(type, field_name)` to their type
            (character), following Python's struct convention.
        """
        pass

    @property
    def field_offsets(self) -> dict[tuple[str, str], int]:
        if hasattr(self, "_field_offsets"):
            return self._field_offsets
        self.read_header()
        return self._field_offsets

    @property
    def field_types(self) -> dict[tuple[str, str], str]:
        if hasattr(self, "_field_types"):
            return self._field_types
        self.read_header()
        return self._field_types

    @property
    def local_particle_count(self) -> int:
        if hasattr(self, "_local_particle_count"):
            return self._local_particle_count
        self.read_header()
        return self._local_particle_count

    @property
    def header(self) -> dict[str, Any]:
        if hasattr(self, "_header"):
            return self._header
        self.read_header()
        return self._header

    def handle_field(
        self, field: tuple[str, str], data_dict: dict[tuple[str, str], np.ndarray]
    ):
        """
        This function allows custom code to be called to handle special cases,
        such as the particle birth time.

        It updates the `data_dict` dictionary with the new data.

        Parameters
        ----------
        field : tuple[str, str]
            The field name.
        data_dict : dict[tuple[str, str], np.ndarray]
            A dictionary containing the data.

        By default, this function does nothing.
        """
        pass


_default_dtypes: dict[int, str] = {
    1: "c",  # char
    2: "h",  # short,
    4: "f",  # float
    8: "d",  # double
}


class DefaultParticleFileHandler(ParticleFileHandler):
    ptype = "io"
    fname = "part_{iout:05d}.out{icpu:05d}"
    file_descriptor = "part_file_descriptor.txt"
    config_field = "ramses-particles"
    reader = _ramses_particle_binary_file_handler

    attrs = (
        ("ncpu", 1, "i"),
        ("ndim", 1, "i"),
        ("npart", 1, "i"),
        ("localseed", -1, "i"),
        ("nstar_tot", 1, "i"),
        ("mstar_tot", 1, "d"),
        ("mstar_lost", 1, "d"),
        ("nsink", 1, "i"),
    )

    known_fields = [
        ("particle_position_x", "d"),
        ("particle_position_y", "d"),
        ("particle_position_z", "d"),
        ("particle_velocity_x", "d"),
        ("particle_velocity_y", "d"),
        ("particle_velocity_z", "d"),
        ("particle_mass", "d"),
        ("particle_identity", "i"),
        ("particle_refinement_level", "i"),
    ]

    def read_header(self):
        if not self.exists:
            self._field_offsets = {}
            self._field_types = {}
            self._local_particle_count = 0
            self._header = {}
            return

        flen = os.path.getsize(self.fname)
        with FortranFile(self.fname) as fd:
            hvals = dict(fd.read_attrs(self.attrs))
            particle_field_pos = fd.tell()

        self._header = hvals
        self._local_particle_count = hvals["npart"]
        extra_particle_fields = self.ds._extra_particle_fields

        if self.has_descriptor:
            particle_fields = _read_part_binary_file_descriptor(self.file_descriptor)
        else:
            particle_fields = list(self.known_fields)

            if extra_particle_fields is not None:
                particle_fields += extra_particle_fields

        if (
            hvals["nstar_tot"] > 0
            and extra_particle_fields is not None
            and ("particle_birth_time", "d") not in particle_fields
        ):
            particle_fields += [
                ("particle_birth_time", "d"),
                ("particle_metallicity", "d"),
            ]

        def build_iterator():
            return chain(
                particle_fields,
                ((f"particle_extra_field_{i}", "d") for i in count(1)),
            )

        field_offsets = {}
        _pfields = {}
        ptype = self.ptype
        blockLen = struct.calcsize("i") * 2

        particle_fields_iterator = build_iterator()
        ipos = particle_field_pos
        while ipos < flen:
            field, vtype = next(particle_fields_iterator)
            field_offsets[ptype, field] = ipos
            _pfields[ptype, field] = vtype
            ipos += blockLen + struct.calcsize(vtype) * hvals["npart"]

        if ipos != flen:
            particle_fields_iterator = build_iterator()
            with FortranFile(self.fname) as fd:
                fd.seek(particle_field_pos)
                ipos = particle_field_pos
                while ipos < flen:
                    field, vtype = next(particle_fields_iterator)
                    old_pos = fd.tell()
                    field_offsets[ptype, field] = old_pos
                    fd.skip(1)
                    ipos = fd.tell()

                    record_len = ipos - old_pos - blockLen
                    exp_len = struct.calcsize(vtype) * hvals["npart"]

                    if record_len != exp_len:
                        # Guess record vtype from length
                        nbytes = record_len // hvals["npart"]
                        # NOTE: in some simulations (e.g. New Horizon), the record length is not
                        # a multiple of 1, 2, 4 or 8. In this case, fallback onto assuming
                        # double precision.
                        vtype = _default_dtypes.get(nbytes, "d")

                        mylog.warning(
                            "Supposed that `%s` has type %s given record size",
                            field,
                            np.dtype(vtype),
                        )

                    _pfields[ptype, field] = vtype

        if field.startswith("particle_extra_field_"):
            iextra = int(field.split("_")[-1])
        else:
            iextra = 0
        if iextra > 0 and not self.ds._warned_extra_fields["io"]:
            mylog.warning(
                "Detected %s extra particle fields assuming kind "
                "`double`. Consider using the `extra_particle_fields` "
                "keyword argument if you have unexpected behavior.",
                iextra,
            )
            self.ds._warned_extra_fields["io"] = True

        if (
            self.ds.use_conformal_time
            and (ptype, "particle_birth_time") in field_offsets
        ):
            field_offsets[ptype, "conformal_birth_time"] = field_offsets[
                ptype, "particle_birth_time"
            ]
            _pfields[ptype, "conformal_birth_time"] = _pfields[
                ptype, "particle_birth_time"
            ]

        self._field_offsets = field_offsets
        self._field_types = _pfields

    @property
    def birth_file_fname(self):
        basename = os.path.abspath(self.ds.root_folder)
        iout = int(
            os.path.basename(self.ds.parameter_filename).split(".")[0].split("_")[1]
        )
        icpu = self.domain_id

        fname = os.path.join(basename, f"birth_{iout:05d}.out{icpu:05d}")
        return fname

    @cached_property
    def has_birth_file(self):
        return os.path.exists(self.birth_file_fname)

    def handle_field(
        self, field: tuple[str, str], data_dict: dict[tuple[str, str], np.ndarray]
    ):
        _ptype, fname = field
        if not (fname == "particle_birth_time" and self.ds.cosmological_simulation):
            return

        # If the birth files exist, read from them
        if self.has_birth_file:
            with FortranFile(self.birth_file_fname) as fd:
                # Note: values are written in Gyr, so we need to convert back to code_time
                data_dict[field] = (
                    self.ds.arr(fd.read_vector("d"), "Gyr").to("code_time").v
                )

            return

        # Otherwise, convert conformal time to physical age
        ds = self.ds
        conformal_time = data_dict[field]
        physical_time = (
            convert_ramses_conformal_time_to_physical_time(ds, conformal_time)
            .to("code_time")
            .v
        )
        # arbitrarily set particles with zero conformal_age to zero
        # particle_age. This corresponds to DM particles.
        data_dict[field] = np.where(conformal_time != 0, physical_time, 0)


class SinkParticleFileHandler(ParticleFileHandler):
    """Handle sink files"""

    ptype = "sink"
    fname = "sink_{iout:05d}.out{icpu:05d}"
    file_descriptor = "sink_file_descriptor.txt"
    config_field = "ramses-sink-particles"
    reader = _ramses_particle_binary_file_handler

    attrs = (("nsink", 1, "i"), ("nindsink", 1, "i"))

    known_fields = [
        ("particle_identifier", "i"),
        ("particle_mass", "d"),
        ("particle_position_x", "d"),
        ("particle_position_y", "d"),
        ("particle_position_z", "d"),
        ("particle_velocity_x", "d"),
        ("particle_velocity_y", "d"),
        ("particle_velocity_z", "d"),
        ("particle_birth_time", "d"),
        ("BH_real_accretion", "d"),
        ("BH_bondi_accretion", "d"),
        ("BH_eddington_accretion", "d"),
        ("BH_esave", "d"),
        ("gas_spin_x", "d"),
        ("gas_spin_y", "d"),
        ("gas_spin_z", "d"),
        ("BH_spin_x", "d"),
        ("BH_spin_y", "d"),
        ("BH_spin_z", "d"),
        ("BH_spin", "d"),
        ("BH_efficiency", "d"),
    ]

    def read_header(self):
        if not self.exists:
            self._field_offsets = {}
            self._field_types = {}
            self._local_particle_count = 0
            self._header = {}
            return
        fd = FortranFile(self.fname)
        flen = os.path.getsize(self.fname)
        hvals = {}
        # Read the header of the file
        attrs = self.attrs

        hvals.update(fd.read_attrs(attrs))
        self._header = hvals

        # This is somehow a trick here: we only want one domain to
        # be read, as ramses writes all the sinks in all the
        # domains. Here, we set the local_particle_count to 0 except
        # for the first domain to be red.
        if getattr(self.ds, "_sink_file_flag", False):
            self._local_particle_count = 0
        else:
            self.ds._sink_file_flag = True
            self._local_particle_count = hvals["nsink"]

        # Read the fields + add the sink properties
        if self.has_descriptor:
            fields = _read_part_binary_file_descriptor(self.file_descriptor)
        else:
            fields = list(self.known_fields)

        # Note: this follows RAMSES convention.
        for i in range(self.ds.dimensionality * 2 + 1):
            for ilvl in range(self.ds.max_level + 1):
                fields.append((f"particle_prop_{ilvl}_{i}", "d"))

        field_offsets = {}
        _pfields = {}

        # Fill the fields, offsets and types
        self.fields = []
        for field, vtype in fields:
            self.fields.append(field)
            if fd.tell() >= flen:
                break
            field_offsets[self.ptype, field] = fd.tell()
            _pfields[self.ptype, field] = vtype
            fd.skip(1)
        self._field_offsets = field_offsets
        self._field_types = _pfields
        fd.close()


class SinkParticleFileHandlerCsv(ParticleFileHandler):
    """Handle sink files from a csv file, the format from the sink particle in ramses"""

    ptype = "sink_csv"
    fname = "sink_{iout:05d}.csv"
    file_descriptor = None
    config_field = "ramses-sink-particles"
    reader = _ramses_particle_csv_file_handler
    attrs = (("nsink", 1, "i"), ("nindsink", 1, "i"))

    def read_header(self):
        if not self.exists:
            self._field_offsets = {}
            self._field_types = {}
            self._local_particle_count = 0
            self._header = {}
            return
        field_offsets = {}
        _pfields = {}

        fields, self._local_particle_count = _read_part_csv_file_descriptor(self.fname)

        for ind, field in enumerate(fields):
            field_offsets[self.ptype, field] = ind
            _pfields[self.ptype, field] = "d"

        self._field_offsets = field_offsets
        self._field_types = _pfields

    def handle_field(
        self, field: tuple[str, str], data_dict: dict[tuple[str, str], np.ndarray]
    ):
        _ptype, fname = field
        if not (fname == "particle_birth_time" and self.ds.cosmological_simulation):
            return

        # convert conformal time to physical age
        ds = self.ds
        conformal_time = data_dict[field]
        physical_time = convert_ramses_conformal_time_to_physical_time(
            ds, conformal_time
        )
        # arbitrarily set particles with zero conformal_age to zero
        # particle_age. This corresponds to DM particles.
        data_dict[field] = np.where(conformal_time > 0, physical_time, 0)



import numpy as np

from .data_structures import GadgetBinaryHeader, GadgetDataset
from .definitions import gadget_field_specs, gadget_ptype_specs
from .io import IOHandlerGadgetBinary

vector_fields = dict(IOHandlerGadgetBinary._vector_fields)

block_ids = {
    "Coordinates": "POS",
    "Velocities": "VEL",
    "ParticleIDs": "ID",
    "Mass": "MASS",
    "InternalEnergy": "U",
    "Density": "RHO",
    "SmoothingLength": "HSML",
}


def write_record(fp, data, endian):
    dtype = endian + "i4"
    size = np.array(data.nbytes, dtype=dtype)
    fp.write(size.tobytes())
    fp.write(data.tobytes())
    fp.write(size.tobytes())


def write_block(fp, data, endian, fmt, block_id):
    assert fmt in [1, 2]
    block_id = "%-4s" % block_id
    if fmt == 2:
        block_id_dtype = np.dtype([("id", "S", 4), ("offset", endian + "i4")])
        block_id_data = np.zeros(1, dtype=block_id_dtype)
        block_id_data["id"] = block_id
        block_id_data["offset"] = data.nbytes + 8
        write_record(fp, block_id_data, endian)
    write_record(fp, data, endian)


def fake_gadget_binary(
    filename="fake_gadget_binary",
    npart=(100, 100, 100, 0, 100, 0),
    header_spec="default",
    field_spec="default",
    ptype_spec="default",
    endian="",
    fmt=2,
):
    """Generate a fake Gadget binary snapshot."""
    header = GadgetBinaryHeader(filename, header_spec)
    field_spec = GadgetDataset._setup_binary_spec(field_spec, gadget_field_specs)
    ptype_spec = GadgetDataset._setup_binary_spec(ptype_spec, gadget_ptype_specs)
    with open(filename, "wb") as fp:
        # Generate and write header blocks
        for i_header, header_spec in enumerate(header.spec):
            specs = []
            for name, dim, dtype in header_spec:
                # workaround a FutureWarning in numpy where np.dtype(name, type, 1)
                # will change meaning in a future version so
                name_dtype = [name, endian + dtype, dim]
                if dim == 1:
                    name_dtype.pop()
                specs.append(tuple(name_dtype))

            header_dtype = np.dtype(specs)
            header = np.zeros(1, dtype=header_dtype)
            if i_header == 0:
                header["Npart"] = npart
                header["Nall"] = npart
                header["NumFiles"] = 1
                header["BoxSize"] = 1
                header["HubbleParam"] = 1
            write_block(fp, header, endian, fmt, "HEAD")

        npart = dict(zip(ptype_spec, npart, strict=True))
        for fs in field_spec:
            # Parse field name and particle type
            if isinstance(fs, str):
                field = fs
                ptype = ptype_spec
            else:
                field, ptype = fs
                if isinstance(ptype, str):
                    ptype = (ptype,)
            # Determine field dimension
            if field in vector_fields:
                dim = vector_fields[field]
            else:
                dim = 1
            # Determine dtype (in numpy convention)
            if field == "ParticleIDs":
                dtype = "u4"
            else:
                dtype = "f4"
            dtype = endian + dtype
            # Generate and write field block
            data = []
            rng = np.random.default_rng()
            for pt in ptype:
                data += [rng.random((npart[pt], dim))]
            data = np.concatenate(data).astype(dtype)
            if field in block_ids:
                block_id = block_ids[field]
            else:
                block_id = ""
            write_block(fp, data, endian, fmt, block_id)
    return filename



import numpy as np

from yt.funcs import mylog

from .parallel_analysis_interface import (
    ResultsStorage,
    _get_comm,
    communication_system,
    parallel_capable,
)

messages = {
    "task": {"msg": "next"},
    "result": {"msg": "result"},
    "task_req": {"msg": "task_req"},
    "end": {"msg": "no_more_tasks"},
}


class TaskQueueNonRoot:
    def __init__(self, tasks, comm, subcomm):
        self.tasks = tasks
        self.results = {}
        self.comm = comm
        self.subcomm = subcomm

    def send_result(self, result):
        new_msg = messages["result"].copy()
        new_msg["value"] = result
        if self.subcomm.rank == 0:
            self.comm.comm.send(new_msg, dest=0, tag=1)
        self.subcomm.barrier()

    def __next__(self):
        msg = messages["task_req"].copy()
        if self.subcomm.rank == 0:
            self.comm.comm.send(msg, dest=0, tag=1)
            msg = self.comm.comm.recv(source=0, tag=2)
        msg = self.subcomm.bcast(msg, root=0)
        if msg["msg"] == messages["end"]["msg"]:
            mylog.debug("Notified to end")
            raise StopIteration
        return msg["value"]

    # For Python 2 compatibility
    next = __next__

    def __iter__(self):
        return self

    def run(self, callable):
        for task in self:
            self.send_result(callable(task))
        return self.finalize()

    def finalize(self, vals=None):
        return self.comm.comm.bcast(vals, root=0)


class TaskQueueRoot(TaskQueueNonRoot):
    def __init__(self, tasks, comm, njobs):
        self.njobs = njobs
        self.tasks = tasks
        self.results = {}
        self.assignments = {}
        self._notified = 0
        self._current = 0
        self._remaining = len(self.tasks)
        self.comm = comm
        # Set up threading here
        # self.dist = threading.Thread(target=self.handle_assignments)
        # self.dist.daemon = True
        # self.dist.start()

    def run(self, func=None):
        self.comm.probe_loop(1, self.handle_assignment)
        return self.finalize(self.results)

    def insert_result(self, source_id, rstore):
        task_id = rstore.result_id
        self.results[task_id] = rstore.result

    def assign_task(self, source_id):
        if self._remaining == 0:
            mylog.debug("Notifying %s to end", source_id)
            msg = messages["end"].copy()
            self._notified += 1
        else:
            msg = messages["task"].copy()
            task_id = self._current
            task = self.tasks[task_id]
            self.assignments[source_id] = task_id
            self._current += 1
            self._remaining -= 1
            msg["value"] = task
        self.comm.comm.send(msg, dest=source_id, tag=2)

    def handle_assignment(self, status):
        msg = self.comm.comm.recv(source=status.source, tag=1)
        if msg["msg"] == messages["result"]["msg"]:
            self.insert_result(status.source, msg["value"])
        elif msg["msg"] == messages["task_req"]["msg"]:
            self.assign_task(status.source)
        else:
            mylog.error("GOT AN UNKNOWN MESSAGE: %s", msg)
            raise RuntimeError
        if self._notified >= self.njobs:
            raise StopIteration


def task_queue(func, tasks, njobs=0):
    comm = _get_comm(())
    if not parallel_capable:
        mylog.error("Cannot create task queue for serial process.")
        raise RuntimeError
    my_size = comm.comm.size
    if njobs <= 0:
        njobs = my_size - 1
    if njobs >= my_size:
        mylog.error(
            "You have asked for %s jobs, but only %s processors are available.",
            njobs,
            (my_size - 1),
        )
        raise RuntimeError
    my_rank = comm.rank
    all_new_comms = np.array_split(np.arange(1, my_size), njobs)
    all_new_comms.insert(0, np.array([0]))
    for i, comm_set in enumerate(all_new_comms):
        if my_rank in comm_set:
            my_new_id = i
            break
    subcomm = communication_system.push_with_ids(all_new_comms[my_new_id].tolist())

    if comm.comm.rank == 0:
        my_q = TaskQueueRoot(tasks, comm, njobs)
    else:
        my_q = TaskQueueNonRoot(None, comm, subcomm)
    communication_system.pop()
    return my_q.run(func)


def dynamic_parallel_objects(tasks, njobs=0, storage=None, broadcast=True):
    comm = _get_comm(())
    if not parallel_capable:
        mylog.error("Cannot create task queue for serial process.")
        raise RuntimeError
    my_size = comm.comm.size
    if njobs <= 0:
        njobs = my_size - 1
    if njobs >= my_size:
        mylog.error(
            "You have asked for %s jobs, but only %s processors are available.",
            njobs,
            (my_size - 1),
        )
        raise RuntimeError
    my_rank = comm.rank
    all_new_comms = np.array_split(np.arange(1, my_size), njobs)
    all_new_comms.insert(0, np.array([0]))
    for i, comm_set in enumerate(all_new_comms):
        if my_rank in comm_set:
            my_new_id = i
            break
    subcomm = communication_system.push_with_ids(all_new_comms[my_new_id].tolist())

    if comm.comm.rank == 0:
        my_q = TaskQueueRoot(tasks, comm, njobs)
        my_q.comm.probe_loop(1, my_q.handle_assignment)
    else:
        my_q = TaskQueueNonRoot(None, comm, subcomm)
        if storage is None:
            for task in my_q:
                yield task
        else:
            for task in my_q:
                rstore = ResultsStorage()
                yield rstore, task
                my_q.send_result(rstore)

    if storage is not None:
        if broadcast:
            my_results = my_q.comm.comm.bcast(my_q.results, root=0)
        else:
            my_results = my_q.results
        storage.update(my_results)

    communication_system.pop()



from yt._typing import KnownFieldsT
from yt.fields.field_info_container import FieldInfoContainer
from yt.utilities.physical_constants import kboltz, mh

b_units = "code_magnetic"
pres_units = "code_pressure"
erg_units = "code_mass * (code_length/code_time)**2"
rho_units = "code_mass / code_length**3"


def velocity_field(comp):
    def _velocity(field, data):
        return data["athena", f"momentum_{comp}"] / data["athena", "density"]

    return _velocity


class AthenaFieldInfo(FieldInfoContainer):
    known_other_fields: KnownFieldsT = (
        ("density", ("code_mass/code_length**3", ["density"], None)),
        ("cell_centered_B_x", (b_units, [], None)),
        ("cell_centered_B_y", (b_units, [], None)),
        ("cell_centered_B_z", (b_units, [], None)),
        ("total_energy", ("code_pressure", ["total_energy_density"], None)),
        (
            "gravitational_potential",
            ("code_velocity**2", ["gravitational_potential"], None),
        ),
    )

    # In Athena, conservative or primitive variables may be written out.
    # By default, yt concerns itself with primitive variables. The following
    # field definitions allow for conversions to primitive variables in the
    # case that the file contains the conservative ones.

    def setup_fluid_fields(self):
        from yt.fields.magnetic_field import setup_magnetic_field_aliases

        unit_system = self.ds.unit_system
        # Add velocity fields
        for comp in "xyz":
            vel_field = ("athena", f"velocity_{comp}")
            mom_field = ("athena", f"momentum_{comp}")
            if vel_field in self.field_list:
                self.add_output_field(
                    vel_field, sampling_type="cell", units="code_length/code_time"
                )
                self.alias(
                    ("gas", f"velocity_{comp}"),
                    vel_field,
                    units=unit_system["velocity"],
                )
            elif mom_field in self.field_list:
                self.add_output_field(
                    mom_field,
                    sampling_type="cell",
                    units="code_mass/code_time/code_length**2",
                )
                self.alias(
                    ("gas", f"momentum_density_{comp}"),
                    mom_field,
                    units=unit_system["density"] * unit_system["velocity"],
                )
                self.add_field(
                    ("gas", f"velocity_{comp}"),
                    sampling_type="cell",
                    function=velocity_field(comp),
                    units=unit_system["velocity"],
                )

        # Add pressure, energy, and temperature fields
        def eint_from_etot(data):
            eint = (
                data["athena", "total_energy"] - data["gas", "kinetic_energy_density"]
            )
            if ("athena", "cell_centered_B_x") in self.field_list:
                eint -= data["gas", "magnetic_energy_density"]
            return eint

        def etot_from_pres(data):
            etot = data["athena", "pressure"] / (data.ds.gamma - 1.0)
            etot += data["gas", "kinetic_energy_density"]
            if ("athena", "cell_centered_B_x") in self.field_list:
                etot += data["gas", "magnetic_energy_density"]
            return etot

        if ("athena", "pressure") in self.field_list:
            self.add_output_field(
                ("athena", "pressure"), sampling_type="cell", units=pres_units
            )
            self.alias(
                ("gas", "pressure"),
                ("athena", "pressure"),
                units=unit_system["pressure"],
            )

            def _specific_thermal_energy(field, data):
                return (
                    data["athena", "pressure"]
                    / (data.ds.gamma - 1.0)
                    / data["athena", "density"]
                )

            self.add_field(
                ("gas", "specific_thermal_energy"),
                sampling_type="cell",
                function=_specific_thermal_energy,
                units=unit_system["specific_energy"],
            )

            def _specific_total_energy(field, data):
                return etot_from_pres(data) / data["athena", "density"]

            self.add_field(
                ("gas", "specific_total_energy"),
                sampling_type="cell",
                function=_specific_total_energy,
                units=unit_system["specific_energy"],
            )
        elif ("athena", "total_energy") in self.field_list:
            self.add_output_field(
                ("athena", "total_energy"), sampling_type="cell", units=pres_units
            )

            def _specific_thermal_energy(field, data):
                return eint_from_etot(data) / data["athena", "density"]

            self.add_field(
                ("gas", "specific_thermal_energy"),
                sampling_type="cell",
                function=_specific_thermal_energy,
                units=unit_system["specific_energy"],
            )

            def _specific_total_energy(field, data):
                return data["athena", "total_energy"] / data["athena", "density"]

            self.add_field(
                ("gas", "specific_total_energy"),
                sampling_type="cell",
                function=_specific_total_energy,
                units=unit_system["specific_energy"],
            )

        # Add temperature field
        def _temperature(field, data):
            return (
                data.ds.mu
                * data["gas", "pressure"]
                / data["gas", "density"]
                * mh
                / kboltz
            )

        self.add_field(
            ("gas", "temperature"),
            sampling_type="cell",
            function=_temperature,
            units=unit_system["temperature"],
        )

        setup_magnetic_field_aliases(
            self, "athena", [f"cell_centered_B_{ax}" for ax in "xyz"]
        )



import os
import shutil
import tempfile

import numpy as np
from nose.tools import assert_raises
from numpy.testing import assert_equal

from yt.data_objects.particle_filters import add_particle_filter, particle_filter
from yt.testing import fake_random_ds, fake_sph_grid_ds
from yt.utilities.exceptions import YTIllDefinedFilter, YTIllDefinedParticleFilter
from yt.visualization.plot_window import ProjectionPlot


def test_add_particle_filter():
    """Test particle filters created via add_particle_filter

    This accesses a deposition field using the particle filter, which was a
    problem in previous versions on this dataset because there are chunks with
    no stars in them.

    """

    def stars(pfilter, data):
        filter_field = (pfilter.filtered_type, "particle_mass")
        return data[filter_field] > 0.5

    add_particle_filter(
        "stars1", function=stars, filtered_type="all", requires=["particle_mass"]
    )
    ds = fake_random_ds(16, nprocs=8, particles=16)
    ds.add_particle_filter("stars1")
    assert ("deposit", "stars1_cic") in ds.derived_field_list

    # Test without requires field
    add_particle_filter("stars2", function=stars)
    ds = fake_random_ds(16, nprocs=8, particles=16)
    ds.add_particle_filter("stars2")
    assert ("deposit", "stars2_cic") in ds.derived_field_list

    # Test adding filter with fields not defined on the ds
    with assert_raises(YTIllDefinedParticleFilter) as ex:
        add_particle_filter(
            "bad_stars", function=stars, filtered_type="all", requires=["wrong_field"]
        )
        ds.add_particle_filter("bad_stars")
    actual = str(ex.exception)
    desired = (
        "\nThe fields\n\t('all', 'wrong_field'),\nrequired by the"
        ' "bad_stars" particle filter, are not defined for this dataset.'
    )
    assert_equal(actual, desired)


def test_add_particle_filter_overriding():
    """Test the add_particle_filter overriding"""
    from yt.data_objects.particle_filters import filter_registry
    from yt.funcs import mylog

    def star_0(pfilter, data):
        pass

    def star_1(pfilter, data):
        pass

    # Use a closure to store whether the warning was called
    def closure(status):
        def warning_patch(*args, **kwargs):
            status[0] = True

        def was_called():
            return status[0]

        return warning_patch, was_called

    ## Test 1: we add a dummy particle filter
    add_particle_filter(
        "dummy", function=star_0, filtered_type="all", requires=["creation_time"]
    )
    assert "dummy" in filter_registry
    assert_equal(filter_registry["dummy"].function, star_0)

    ## Test 2: we add another dummy particle filter.
    ##         a warning is expected. We use the above closure to
    ##         check that.
    # Store the original warning function
    warning = mylog.warning
    monkey_warning, monkey_patch_was_called = closure([False])
    mylog.warning = monkey_warning
    add_particle_filter(
        "dummy", function=star_1, filtered_type="all", requires=["creation_time"]
    )
    assert_equal(filter_registry["dummy"].function, star_1)
    assert_equal(monkey_patch_was_called(), True)

    # Restore the original warning function
    mylog.warning = warning


def test_particle_filter_decorator():
    """Test the particle_filter decorator"""

    @particle_filter(filtered_type="all", requires=["particle_mass"])
    def heavy_stars(pfilter, data):
        filter_field = (pfilter.filtered_type, "particle_mass")
        return data[filter_field] > 0.5

    ds = fake_random_ds(16, nprocs=8, particles=16)
    ds.add_particle_filter("heavy_stars")
    assert "heavy_stars" in ds.particle_types
    assert ("deposit", "heavy_stars_cic") in ds.derived_field_list

    # Test name of particle filter
    @particle_filter(name="my_stars", filtered_type="all", requires=["particle_mass"])
    def custom_stars(pfilter, data):
        filter_field = (pfilter.filtered_type, "particle_mass")
        return data[filter_field] == 0.5

    ds = fake_random_ds(16, nprocs=8, particles=16)
    ds.add_particle_filter("my_stars")
    assert "my_stars" in ds.particle_types
    assert ("deposit", "my_stars_cic") in ds.derived_field_list


def test_particle_filter_exceptions():
    @particle_filter(filtered_type="all", requires=["particle_mass"])
    def filter1(pfilter, data):
        return data

    ds = fake_random_ds(16, nprocs=8, particles=16)
    ds.add_particle_filter("filter1")

    ad = ds.all_data()
    with assert_raises(YTIllDefinedFilter):
        ad["filter1", "particle_mass"].shape[0]

    @particle_filter(filtered_type="all", requires=["particle_mass"])
    def filter2(pfilter, data):
        filter_field = ("io", "particle_mass")
        return data[filter_field] > 0.5

    ds.add_particle_filter("filter2")
    ad = ds.all_data()
    ad["filter2", "particle_mass"].min()


def test_particle_filter_dependency():
    """
    Test dataset add_particle_filter which should automatically add
    the dependency of the filter.
    """

    @particle_filter(filtered_type="all", requires=["particle_mass"])
    def h_stars(pfilter, data):
        filter_field = (pfilter.filtered_type, "particle_mass")
        return data[filter_field] > 0.5

    @particle_filter(filtered_type="h_stars", requires=["particle_mass"])
    def hh_stars(pfilter, data):
        filter_field = (pfilter.filtered_type, "particle_mass")
        return data[filter_field] > 0.9

    ds = fake_random_ds(16, nprocs=8, particles=16)
    ds.add_particle_filter("hh_stars")
    assert "hh_stars" in ds.particle_types
    assert "h_stars" in ds.particle_types
    assert ("deposit", "hh_stars_cic") in ds.derived_field_list
    assert ("deposit", "h_stars_cic") in ds.derived_field_list


def test_covering_grid_particle_filter():
    @particle_filter(filtered_type="all", requires=["particle_mass"])
    def heavy_stars(pfilter, data):
        filter_field = (pfilter.filtered_type, "particle_mass")
        return data[filter_field] > 0.5

    ds = fake_random_ds(16, nprocs=8, particles=16)
    ds.add_particle_filter("heavy_stars")

    for grid in ds.index.grids:
        cg = ds.covering_grid(grid.Level, grid.LeftEdge, grid.ActiveDimensions)

        assert_equal(
            cg["heavy_stars", "particle_mass"].shape[0],
            grid["heavy_stars", "particle_mass"].shape[0],
        )
        assert_equal(
            cg["heavy_stars", "particle_mass"].shape[0],
            grid["heavy_stars", "particle_mass"].shape[0],
        )


def test_sph_particle_filter_plotting():
    ds = fake_sph_grid_ds()

    @particle_filter("central_gas", requires=["particle_position"], filtered_type="io")
    def _filter(pfilter, data):
        coords = np.abs(data[pfilter.filtered_type, "particle_position"])
        return (coords[:, 0] < 1.6) & (coords[:, 1] < 1.6) & (coords[:, 2] < 1.6)

    ds.add_particle_filter("central_gas")

    plot = ProjectionPlot(ds, "z", ("central_gas", "density"))
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    plot.save()

    os.chdir(curdir)
    shutil.rmtree(tmpdir)



import numpy as np

from yt.funcs import mylog
from yt.utilities.io_handler import BaseIOHandler

from .data_structures import chk23

float_size = {"float": np.dtype(">f4").itemsize, "double": np.dtype(">f8").itemsize}

axis_list = ["_x", "_y", "_z"]


class IOHandlerAthena(BaseIOHandler):
    _dataset_type = "athena"
    _offset_string = "data:offsets=0"
    _data_string = "data:datatype=0"
    _read_table_offset = None

    def _read_field_names(self, grid):
        pass

    def _read_chunk_data(self, chunk, fields):
        data = {}
        if len(chunk.objs) == 0:
            return data
        for grid in chunk.objs:
            if grid.filename is None:
                continue
            f = open(grid.filename, "rb")
            data[grid.id] = {}
            grid_dims = grid.ActiveDimensions
            read_dims = grid.read_dims.astype("int64")
            grid_ncells = np.prod(read_dims)
            grid0_ncells = np.prod(grid.index.grids[0].read_dims)
            read_table_offset = get_read_table_offset(f)
            for field in fields:
                ftype, offsetr, dtype = grid.index._field_map[field]
                if grid_ncells != grid0_ncells:
                    offset = offsetr + (
                        (grid_ncells - grid0_ncells) * (offsetr // grid0_ncells)
                    )
                if grid_ncells == grid0_ncells:
                    offset = offsetr
                offset = int(offset)  # Casting to be certain.
                file_offset = (
                    grid.file_offset[2]
                    * read_dims[0]
                    * read_dims[1]
                    * float_size[dtype]
                )
                xread = slice(grid.file_offset[0], grid.file_offset[0] + grid_dims[0])
                yread = slice(grid.file_offset[1], grid.file_offset[1] + grid_dims[1])
                f.seek(read_table_offset + offset + file_offset)
                if dtype == "float":
                    dt = ">f4"
                elif dtype == "double":
                    dt = ">f8"
                if ftype == "scalar":
                    f.seek(read_table_offset + offset + file_offset)
                    v = np.fromfile(f, dtype=dt, count=grid_ncells).reshape(
                        read_dims, order="F"
                    )
                if ftype == "vector":
                    vec_offset = axis_list.index(field[-1][-2:])
                    f.seek(read_table_offset + offset + 3 * file_offset)
                    v = np.fromfile(f, dtype=dt, count=3 * grid_ncells)
                    v = v[vec_offset::3].reshape(read_dims, order="F")
                if grid.ds.field_ordering == 1:
                    data[grid.id][field] = v[xread, yread, :].T.astype("float64")
                else:
                    data[grid.id][field] = v[xread, yread, :].astype("float64")
            f.close()
        return data

    def _read_data_slice(self, grid, field, axis, coord):
        sl = [slice(None), slice(None), slice(None)]
        sl[axis] = slice(coord, coord + 1)
        if grid.ds.field_ordering == 1:
            sl.reverse()
        return self._read_data_set(grid, field)[tuple(sl)]

    def _read_fluid_selection(self, chunks, selector, fields, size):
        chunks = list(chunks)
        if any((ftype != "athena" for ftype, fname in fields)):
            raise NotImplementedError
        rv = {}
        for field in fields:
            rv[field] = np.empty(size, dtype="float64")
        ng = sum(len(c.objs) for c in chunks)
        mylog.debug(
            "Reading %s cells of %s fields in %s grids",
            size,
            [f2 for f1, f2 in fields],
            ng,
        )
        ind = 0
        for chunk in chunks:
            data = self._read_chunk_data(chunk, fields)
            for g in chunk.objs:
                for field in fields:
                    ftype, fname = field
                    ds = data[g.id].pop(field)
                    nd = g.select(selector, ds, rv[field], ind)  # caches
                ind += nd
                data.pop(g.id)
        return rv


def get_read_table_offset(f):
    line = f.readline()
    while True:
        splitup = line.strip().split()
        chkc = chk23("CELL_DATA")
        chkp = chk23("POINT_DATA")
        if chkc in splitup or chkp in splitup:
            f.readline()
            read_table_offset = f.tell()
            break
        line = f.readline()
    return read_table_offset



# distutils: language = c++
# distutils: libraries = STD_LIBS

cimport numpy as np

import numpy as np

from libc.math cimport sqrt
import cython



@cython.cdivision(True)
cdef inline double _a_dot(double a, double h0, double om_m, double om_l) noexcept:
    om_k = 1.0 - om_m - om_l
    return h0 * a * sqrt(om_m * (a ** -3) + om_k * (a ** -2) + om_l)


@cython.cdivision(True)
cpdef double _a_dot_recip(double a, double h0, double om_m, double om_l):
    return 1. / _a_dot(a, h0, om_m, om_l)


cdef inline double _da_dtau(double a, double h0, double om_m, double om_l) noexcept:
    return a**2 * _a_dot(a, h0, om_m, om_l)

@cython.cdivision(True)
cpdef double _da_dtau_recip(double a, double h0, double om_m, double om_l) noexcept:
    return 1. / _da_dtau(a, h0, om_m, om_l)



def t_frw(ds, z):
    from scipy.integrate import quad
    aexp = 1 / (1 + z)

    h0 = ds.hubble_constant
    om_m = ds.omega_matter
    om_l = ds.omega_lambda
    conv = ds.quan(0.01, "Mpc/km*s").to("Gyr")

    if isinstance(z, (int, float)):
        return ds.quan(
            quad(_a_dot_recip, 0, aexp, args=(h0, om_m, om_l))[0],
            units=conv,
        )

    return ds.arr(
        [quad(_a_dot_recip, 0, a, args=(h0, om_m, om_l))[0] for a in aexp],
        units=conv,
    )

def tau_frw(ds, z):
    from scipy.integrate import quad
    aexp = 1 / (1 + z)

    h0 = ds.hubble_constant
    om_m = ds.omega_matter
    om_l = ds.omega_lambda

    if isinstance(z, (int, float)):
        return quad(_da_dtau_recip, 1, aexp, args=(h0, om_m, om_l))[0]

    return np.asarray(
        [quad(_da_dtau_recip, 1, a, args=(h0, om_m, om_l))[0] for a in aexp],
    )



"""
Utilities for line integral convolution annotation



"""



import numpy as np

cimport cython
cimport numpy as np


@cython.cdivision(True)
cdef void _advance_2d(double vx, double vy,
                     int* x, int* y,
                     double* fx, double* fy,
                     int w, int h):
    cdef double tx, ty
    if vx>=0:
        tx = (1-fx[0])/vx
    else:
        tx = -fx[0]/vx
    if vy>=0:
        ty = (1-fy[0])/vy
    else:
        ty = -fy[0]/vy
    if tx<ty:
        if vx>=0:
            x[0]+=1
            fx[0]=0
        else:
            x[0]-=1
            fx[0]=1
        fy[0]+=tx*vy
    else:
        if vy>=0:
            y[0]+=1
            fy[0]=0
        else:
            y[0]-=1
            fy[0]=1
        fx[0]+=ty*vx
    if x[0]>=w:
        x[0]=w-1
    if x[0]<0:
        x[0]=0
    if y[0]<0:
        y[0]=0
    if y[0]>=h:
        y[0]=h-1

def line_integral_convolution_2d(
        np.ndarray[double, ndim=3] vectors,
        np.ndarray[double, ndim=2] texture,
        np.ndarray[double, ndim=1] kernel):
    cdef int i,j,l,x,y
    cdef int h,w,kernellen
    cdef double fx, fy
    cdef np.ndarray[double, ndim=2] result

    w = vectors.shape[0]
    h = vectors.shape[1]

    kernellen = kernel.shape[0]
    result = np.zeros((w,h),dtype=np.double)

    vectors = vectors[...,::-1].copy()

    for i in range(w):
        for j in range(h):
            if vectors[i,j,0]==0 and vectors[i,j,1]==0:
                continue
            x = i
            y = j
            fx = 0.5
            fy = 0.5

            l = kernellen//2
            result[i,j] += kernel[l]*texture[x,y]
            while l<kernellen-1:
                _advance_2d(vectors[x,y,0],vectors[x,y,1],
                        &x, &y, &fx, &fy, w, h)
                l+=1
                result[i,j] += kernel[l]*texture[x,y]

            x = i
            y = j
            fx = 0.5
            fy = 0.5

            while l>0:
                _advance_2d(-vectors[x,y,0],-vectors[x,y,1],
                        &x, &y, &fx, &fy, w, h)
                l-=1
                result[i,j] += kernel[l]*texture[x,y]

    return result



# distutils: libraries = STD_LIBS
"""




"""


import numpy as np

cimport cython
cimport libc.math as math
cimport numpy as np


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def _cyl2cart(np.ndarray[np.float64_t, ndim=2] x):
    """Converts points in cylindrical coordinates to cartesian points."""
    # NOTE this should be removed once the coord interface comes online
    cdef int i, I
    cdef np.ndarray[np.float64_t, ndim=2] xcart
    I = x.shape[0]
    xcart = np.empty((I, x.shape[1]), dtype='float64')
    for i in range(I):
        xcart[i,0] = x[i,0] * math.cos(x[i,2])
        xcart[i,1] = x[i,0] * math.sin(x[i,2])
        xcart[i,2] = x[i,1]
    return xcart

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def _cart_intersect(np.ndarray[np.float64_t, ndim=1] a,
                    np.ndarray[np.float64_t, ndim=1] b,
                    np.ndarray[np.float64_t, ndim=2] c,
                    np.ndarray[np.float64_t, ndim=2] d):
    """Finds the times and locations of the lines defined by a->b and
    c->d in cartesian space.  a and b must be 1d points.  c and d must
    be 2d arrays of equal shape whose second dim is the same length as
    a and b.
    """
    cdef int i, I
    cdef np.ndarray[np.float64_t, ndim=1] r, s, t
    cdef np.ndarray[np.float64_t, ndim=2] loc
    I = c.shape[0]
    shape = (I, c.shape[1])
    t = np.empty(I, dtype='float64')
    loc = np.empty(shape, dtype='float64')

    r = b - a
    for i in range(I):
        s = d[i] - c[i]
        t[i] = (np.cross(c[i] - a, s).sum()) / (np.cross(r, s).sum())
        loc[i] = a + t[i]*r
    return t, loc


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def cylindrical_ray_trace(np.ndarray[np.float64_t, ndim=1] p1,
                          np.ndarray[np.float64_t, ndim=1] p2,
                          np.ndarray[np.float64_t, ndim=2] left_edges,
                          np.ndarray[np.float64_t, ndim=2] right_edges):
    """Computes straight (cartesian) rays being traced through a
    cylindrical geometry.

    Parameters
    ----------
    p1 : length 3 float ndarray
        start point for ray
    p2 : length 3 float ndarray
        stop point for ray
    left_edges : 2d ndarray
        left edges of grid cells
    right_edges : 2d ndarray
        right edges of grid cells

    Returns
    -------
    t : 1d float ndarray
        ray parametric time on range [0,1]
    s : 1d float ndarray
        ray parametric distance on range [0,len(ray)]
    rztheta : 2d float ndarray
        ray grid cell intersections in cylindrical coordinates
    inds : 1d int ndarray
        indexes into the grid cells which the ray crosses in order.

    """
    cdef np.int64_t i, I
    cdef np.float64_t a, b, bsqrd, twoa
    cdef np.ndarray[np.float64_t, ndim=1] p1cart, p2cart, dpcart, t, s, \
                                          rleft, rright, zleft, zright, \
                                          cleft, cright, thetaleft, thetaright, \
                                          tmleft, tpleft, tmright, tpright, tsect
    cdef np.ndarray[np.int64_t, ndim=1, cast=True] inds, tinds, sinds
    cdef np.ndarray[np.float64_t, ndim=2] xyz, rztheta, ptemp, b1, b2, dsect

    # set up  points
    ptemp = np.array([p1, p2])
    ptemp = _cyl2cart(ptemp)
    p1cart = ptemp[0]
    p2cart = ptemp[1]
    dpcart = p2cart - p1cart

    # set up components
    rleft = left_edges[:,0]
    rright = right_edges[:,0]
    zleft = left_edges[:,1]
    zright = right_edges[:,1]

    a = dpcart[0] * dpcart[0] + dpcart[1] * dpcart[1]
    b = 2 * dpcart[0] * p1cart[0] + 2 * dpcart[1] * p1cart[1]
    cleft = p1cart[0] * p1cart[0] + p1cart[1] * p1cart[1] - rleft * rleft
    cright = p1cart[0] * p1cart[0] + p1cart[1] * p1cart[1] - rright * rright
    twoa = 2 * a
    bsqrd = b * b

    # Compute positive and negative times and associated masks
    I = np.intp(left_edges.shape[0])
    tmleft = np.empty(I, dtype='float64')
    tpleft = np.empty(I, dtype='float64')
    tmright = np.empty(I, dtype='float64')
    tpright = np.empty(I, dtype='float64')
    for i in range(I):
        tmleft[i] = (-b - math.sqrt(bsqrd - 4*a*cleft[i])) / twoa
        tpleft[i] = (-b + math.sqrt(bsqrd - 4*a*cleft[i])) / twoa
        tmright[i] = (-b - math.sqrt(bsqrd - 4*a*cright[i])) / twoa
        tpright[i] = (-b + math.sqrt(bsqrd - 4*a*cright[i])) / twoa

    tmmright = np.logical_and(~np.isnan(tmright), rright <= p1[0])
    tpmright = np.logical_and(~np.isnan(tpright), rright <= p2[0])

    tmmleft = np.logical_and(~np.isnan(tmleft), rleft <= p1[0])
    tpmleft = np.logical_and(~np.isnan(tpleft), rleft <= p2[0])

    # compute first cut of indexes and thetas, which
    # have been filtered by those values for which intersection
    # times are impossible (see above masks). Note that this is
    # still independent of z.
    inds = np.unique(np.concatenate([np.argwhere(tmmleft).flat,
                                     np.argwhere(tpmleft).flat,
                                     np.argwhere(tmmright).flat,
                                     np.argwhere(tpmright).flat,]))
    if 0 == inds.shape[0]:
        inds = np.arange(np.intp(I))
        thetaleft = np.empty(I)
        thetaleft.fill(p1[2])
        thetaright = np.empty(I)
        thetaright.fill(p2[2])
    else:
        rleft = rleft[inds]
        rright = rright[inds]

        zleft = zleft[inds]
        zright = zright[inds]

        thetaleft = np.arctan2((p1cart[1] + tmleft[inds]*dpcart[1]),
                               (p1cart[0] + tmleft[inds]*dpcart[0]))
        nans = np.isnan(thetaleft)
        thetaleft[nans] = np.arctan2((p1cart[1] + tpleft[inds[nans]]*dpcart[1]),
                                     (p1cart[0] + tpleft[inds[nans]]*dpcart[0]))
        thetaright = np.arctan2((p1cart[1] + tmright[inds]*dpcart[1]),
                                (p1cart[0] + tmright[inds]*dpcart[0]))
        nans = np.isnan(thetaright)
        thetaright[nans] = np.arctan2((p1cart[1] + tpright[inds[nans]]*dpcart[1]),
                                      (p1cart[0] + tpright[inds[nans]]*dpcart[0]))

    # Set up the cell boundary arrays
    b1 = np.concatenate([[rleft,  zright, thetaleft],
                         [rleft,  zleft,  thetaleft],
                         [rleft,  zleft,  thetaleft],
                         [rright, zleft,  thetaright],
                         [rleft,  zleft,  thetaleft],
                         [rright, zright, thetaleft],
                         [rleft,  zright, thetaleft],
                         [rright, zleft,  thetaleft],
                         ], axis=1).T

    b2 = np.concatenate([[rright, zright, thetaright],
                         [rright, zleft,  thetaright],
                         [rleft,  zright, thetaleft],
                         [rright, zright, thetaright],
                         [rleft,  zleft,  thetaright],
                         [rright, zright, thetaright],
                         [rleft,  zright, thetaright],
                         [rright, zleft,  thetaright],
                         ], axis=1).T

    inds = np.concatenate([inds, inds, inds, inds, inds, inds, inds, inds])

    # find intersections and compute return values
    tsect, dsect = _cart_intersect(p1cart, p2cart, _cyl2cart(b1), _cyl2cart(b2))
    tmask = np.logical_and(0.0<=tsect, tsect<=1.0)
    ret = np.unique(tsect[tmask], return_index=True)
    tsect, tinds = ret[0], ret[1].astype('int64')
    inds = inds[tmask][tinds]
    xyz = dsect[tmask][tinds]
    s = np.sqrt(((xyz - p1cart) * (xyz - p1cart)).sum(axis=1))
    ret = np.unique(s, return_index=True)
    s, sinds = ret[0], ret[1].astype('int64')
    inds = inds[sinds]
    xyz = xyz[sinds]
    t = s/np.sqrt((dpcart*dpcart).sum())
    sinds = s.argsort()
    s = s[sinds]
    t = t[sinds]
    inds = inds[sinds]
    xyz = xyz[sinds]
    rztheta = np.concatenate([np.sqrt(xyz[:,0] * xyz[:,0] + xyz[:,1] * xyz[:,1])[:,np.newaxis],
                              xyz[:,2:3],
                              np.arctan2(xyz[:,1], xyz[:,0])[:,np.newaxis]], axis=1)
    return t, s, rztheta, inds
    #rztheta[:,2] = 0.0 + (rztheta[:,2] - np.pi*3/2)%(2*np.pi)



cimport numpy as np
from libc.stdio cimport FILE

ctypedef np.int32_t INT32_t
ctypedef np.int64_t INT64_t
ctypedef np.float64_t DOUBLE_t

cdef class FortranFile:
    cdef FILE* cfile
    cdef bint _closed

    cpdef INT64_t skip(self, INT64_t n=*) except -1
    cdef INT64_t get_size(self, str dtype)
    cpdef INT32_t read_int(self) except? -1
    cpdef np.ndarray read_vector(self, str dtype)
    cdef int read_vector_inplace(self, str dtype, void *data)
    cpdef INT64_t tell(self) except -1
    cpdef INT64_t seek(self, INT64_t pos, INT64_t whence=*) except -1
    cpdef void close(self)



import numpy as np
from numpy.testing import assert_equal

from yt import load
from yt.testing import (
    assert_rel_equal,
    cubicspline_python,
    fake_random_ds,
    fake_sph_grid_ds,
    integrate_kernel,
    requires_file,
    requires_module,
)
from yt.units._numpy_wrapper_functions import uconcatenate


def test_ray():
    for nproc in [1, 2, 4, 8]:
        ds = fake_random_ds(64, nprocs=nproc)
        dx = (ds.domain_right_edge - ds.domain_left_edge) / ds.domain_dimensions
        # Three we choose, to get varying vectors, and ten random
        pp1 = np.random.random((3, 13))
        pp2 = np.random.random((3, 13))
        pp1[:, 0] = [0.1, 0.2, 0.3]
        pp2[:, 0] = [0.8, 0.1, 0.4]
        pp1[:, 1] = [0.9, 0.2, 0.3]
        pp2[:, 1] = [0.8, 0.1, 0.4]
        pp1[:, 2] = [0.9, 0.2, 0.9]
        pp2[:, 2] = [0.8, 0.1, 0.4]
        unitary = ds.arr(1.0, "")
        for i in range(pp1.shape[1]):
            p1 = ds.arr(pp1[:, i] + 1e-8 * np.random.random(3), "code_length")
            p2 = ds.arr(pp2[:, i] + 1e-8 * np.random.random(3), "code_length")

            my_ray = ds.ray(p1, p2)
            assert_rel_equal(my_ray["dts"].sum(), unitary, 14)
            ray_cells = my_ray["dts"] > 0

            # find cells intersected by the ray
            my_all = ds.all_data()

            dt = np.abs(dx / (p2 - p1))
            tin = uconcatenate(
                [
                    [(my_all["index", "x"] - p1[0]) / (p2 - p1)[0] - 0.5 * dt[0]],
                    [(my_all["index", "y"] - p1[1]) / (p2 - p1)[1] - 0.5 * dt[1]],
                    [(my_all["index", "z"] - p1[2]) / (p2 - p1)[2] - 0.5 * dt[2]],
                ]
            )
            tout = uconcatenate(
                [
                    [(my_all["index", "x"] - p1[0]) / (p2 - p1)[0] + 0.5 * dt[0]],
                    [(my_all["index", "y"] - p1[1]) / (p2 - p1)[1] + 0.5 * dt[1]],
                    [(my_all["index", "z"] - p1[2]) / (p2 - p1)[2] + 0.5 * dt[2]],
                ]
            )
            tin = tin.max(axis=0)
            tout = tout.min(axis=0)
            my_cells = (tin < tout) & (tin < 1) & (tout > 0)

            assert_equal(ray_cells.sum(), my_cells.sum())
            assert_rel_equal(
                my_ray["gas", "density"][ray_cells].sum(),
                my_all["gas", "density"][my_cells].sum(),
                14,
            )
            assert_rel_equal(my_ray["dts"].sum(), unitary, 14)


@requires_module("h5py")
@requires_file("GadgetDiskGalaxy/snapshot_200.hdf5")
def test_ray_particle():
    ds = load("GadgetDiskGalaxy/snapshot_200.hdf5")
    ray = ds.ray(ds.domain_left_edge, ds.domain_right_edge)
    assert_equal(ray["t"].shape, (1451,))
    assert ray["dts"].sum(dtype="f8") > 0


## test that kernels integrate correctly
# (1) including the right particles
# (2) correct t and dts values for those particles
## fake_sph_grid_ds:
# This dataset should have 27 particles with the particles arranged
# uniformly on a 3D grid. The bottom left corner is (0.5,0.5,0.5) and
# the top right corner is (2.5,2.5,2.5). All particles will have
# non-overlapping smoothing regions with a radius of 0.05, masses of
# 1, and densities of 1, and zero velocity.


def test_ray_particle2():
    kernelfunc = cubicspline_python
    ds = fake_sph_grid_ds(hsml_factor=1.0)
    ds.kernel_name = "cubic"

    ## Ray through the one particle at (0.5, 0.5, 0.5):
    ## test basic kernel integration
    eps = 0.0  # 1e-7
    start0 = np.array((1.0 + eps, 0.0, 0.5))
    end0 = np.array((0.0, 1.0 + eps, 0.5))
    ray0 = ds.ray(start0, end0)
    b0 = np.array([np.sqrt(2.0) * eps])
    hsml0 = np.array([0.05])
    len0 = np.sqrt(np.sum((end0 - start0) ** 2))
    # for a ParticleDataset like this one, the Ray object attempts
    # to generate the 't' and 'dts' fields using the grid method
    ray0.field_data["t"] = ray0.ds.arr(ray0._generate_container_field_sph("t"))
    ray0.field_data["dts"] = ray0.ds.arr(ray0._generate_container_field_sph("dts"))
    # not demanding too much precision;
    # from kernel volume integrals, the linear interpolation
    # restricts you to 4 -- 5 digits precision
    assert_equal(ray0["t"].shape, (1,))
    assert_rel_equal(ray0["t"], np.array([0.5]), 5)
    assert_rel_equal(ray0[("gas", "position")].v, np.array([[0.5, 0.5, 0.5]]), 5)
    dl0 = integrate_kernel(kernelfunc, b0, hsml0)
    dl0 *= ray0[("gas", "mass")].v / ray0[("gas", "density")].v
    assert_rel_equal(ray0[("dts")].v, dl0 / len0, 4)

    ## Ray in the middle of the box:
    ## test end points, >1 particle
    start1 = np.array((1.53, 0.53, 1.0))
    end1 = np.array((1.53, 0.53, 3.0))
    ray1 = ds.ray(start1, end1)
    b1 = np.array([np.sqrt(2.0) * 0.03] * 2)
    hsml1 = np.array([0.05] * 2)
    len1 = np.sqrt(np.sum((end1 - start1) ** 2))
    # for a ParticleDataset like this one, the Ray object attempts
    # to generate the 't' and 'dts' fields using the grid method
    ray1.field_data["t"] = ray1.ds.arr(ray1._generate_container_field_sph("t"))
    ray1.field_data["dts"] = ray1.ds.arr(ray1._generate_container_field_sph("dts"))
    # not demanding too much precision;
    # from kernel volume integrals, the linear interpolation
    # restricts you to 4 -- 5 digits precision
    assert_equal(ray1["t"].shape, (2,))
    assert_rel_equal(ray1["t"], np.array([0.25, 0.75]), 5)
    assert_rel_equal(
        ray1[("gas", "position")].v, np.array([[1.5, 0.5, 1.5], [1.5, 0.5, 2.5]]), 5
    )
    dl1 = integrate_kernel(kernelfunc, b1, hsml1)
    dl1 *= ray1[("gas", "mass")].v / ray1[("gas", "density")].v
    assert_rel_equal(ray1[("dts")].v, dl1 / len1, 4)

    ## Ray missing all particles:
    ## test handling of size-0 selections
    start2 = np.array((1.0, 2.0, 0.0))
    end2 = np.array((1.0, 2.0, 3.0))
    ray2 = ds.ray(start2, end2)
    # for a ParticleDataset like this one, the Ray object attempts
    # to generate the 't' and 'dts' fields using the grid method
    ray2.field_data["t"] = ray2.ds.arr(ray2._generate_container_field_sph("t"))
    ray2.field_data["dts"] = ray2.ds.arr(ray2._generate_container_field_sph("dts"))
    assert_equal(ray2["t"].shape, (0,))
    assert_equal(ray2["dts"].shape, (0,))
    assert_equal(ray2[("gas", "position")].v.shape, (0, 3))



# distutils: include_dirs = LIB_DIR
# distutils: libraries = STD_LIBS
"""
Oct visitor functions




"""


cimport cython
cimport numpy as np

import numpy as np

from libc.stdlib cimport malloc

from yt.utilities.lib.fp_utils cimport *
from yt.utilities.lib.geometry_utils cimport encode_morton_64bit

from .oct_container cimport OctreeContainer

# Now some visitor functions

cdef class OctVisitor:
    def __init__(self, OctreeContainer octree, int domain_id = -1):
        cdef int i
        self.index = 0
        self.last = -1
        self.global_index = -1
        for i in range(3):
            self.pos[i] = -1
            self.ind[i] = -1
        self.dims = 0
        self.domain = domain_id
        self.level = -1
        self.nz = octree.nz
        self.nzones = self.nz**3

    cdef void visit(self, Oct* o, np.uint8_t selected):
        raise NotImplementedError

# This copies an integer array from the source to the destination, based on the
# selection criteria.
cdef class CopyArrayI64(OctVisitor):
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        # We should always have global_index less than our source.
        # "last" here tells us the dimensionality of the array.
        if selected == 0: return
        # There are this many records between "octs"
        self.dest[self.index, :] = self.source[
                self.ind[2], self.ind[1], self.ind[0],
                self.global_index, :]
        self.index += 1

# This copies a floating point array from the source to the destination, based
# on the selection criteria.
cdef class CopyArrayF64(OctVisitor):
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        # We should always have global_index less than our source.
        # "last" here tells us the dimensionality of the array.
        if selected == 0: return
        # There are this many records between "octs"
        self.dest[self.index, :] = self.source[
                self.ind[2], self.ind[1], self.ind[0],
                self.global_index, :]
        self.index += 1

# This copies a bit array from source to the destination, based on file_ind
cdef class CopyFileIndArrayI8(OctVisitor):
    def __init__(self, OctreeContainer octree, int domain_id = -1):
        super(CopyFileIndArrayI8, self).__init__(octree, domain_id)
        self.root = -1
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        if self.level == 0:
            self.root += 1
        if self.last != o.domain_ind:
            self.last = o.domain_ind
            self.dest[o.domain_ind] = self.source[self.root]
            self.index += 1

# This counts the number of octs, selected or not, that the selector hits.
# Note that the selector will not recursively visit unselected octs, so this is
# still useful.
cdef class CountTotalOcts(OctVisitor):
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        # Count even if not selected.
        # Number of *octs* visited.
        if self.last != o.domain_ind:
            self.index += 1
            self.last = o.domain_ind

# This counts the number of selected cells.
cdef class CountTotalCells(OctVisitor):
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        # Number of *cells* visited and selected.
        self.index += selected

# Every time a cell is visited, mark it.  This will be for all visited octs.
cdef class MarkOcts(OctVisitor):
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        # We mark them even if they are not selected
        if self.last != o.domain_ind:
            self.last = o.domain_ind
            self.index += 1
        self.mark[self.index, self.ind[2], self.ind[1], self.ind[0]] = 1

# Mask all the selected cells.
cdef class MaskOcts(OctVisitor):
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        if selected == 0: return
        self.mask[self.global_index, self.ind[2], self.ind[1], self.ind[0]] = 1

# Compute a mapping from domain_ind to flattened index.
cdef class IndexOcts(OctVisitor):
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        # Note that we provide an index even if the cell is not selected.
        if self.last != o.domain_ind:
            self.last = o.domain_ind
            self.oct_index[o.domain_ind] = self.index
            self.index += 1

# Compute a mapping from domain_ind to flattened index with some octs masked.
cdef class MaskedIndexOcts(OctVisitor):
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        # Note that we provide an index even if the cell is not selected.
        if self.last != o.domain_ind:
            self.last = o.domain_ind
            if self.oct_mask[o.domain_ind] == 1:
                self.oct_index[o.domain_ind] = self.index
                self.index += 1

# Compute a mapping from domain_ind to flattened index checking mask.
cdef class IndexMaskMapOcts(OctVisitor):
    def __init__(self, OctreeContainer octree, int domain_id = -1):
        super(IndexMaskMapOcts, self).__init__(octree, domain_id)
        self.map_index = 0
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        if self.last != o.domain_ind:
            self.last = o.domain_ind
            if self.oct_mask[o.domain_ind] == 1:
                if self.map_domain_ind[self.map_index] >= 0:
                    self.oct_index[self.map_domain_ind[self.map_index]] = self.index
                self.map_index += 1
            self.index += 1

# Integer coordinates
cdef class ICoordsOcts(OctVisitor):
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        if selected == 0: return
        cdef int i
        for i in range(3):
            self.icoords[self.index,i] = (self.pos[i] * self.nz) + self.ind[i]
        self.index += 1

# Level
cdef class IResOcts(OctVisitor):
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        if selected == 0: return
        self.ires[self.index] = self.level
        self.index += 1

# Floating point coordinates
cdef class FCoordsOcts(OctVisitor):
    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        # Note that this does not actually give the correct floating point
        # coordinates.  It gives them in some unit system where the domain is 1.0
        # in all directions, and assumes that they will be scaled later.
        if selected == 0: return
        cdef int i
        cdef np.float64_t c, dx
        dx = 1.0 / ((self.nz) << self.level)
        for i in range(3):
            c = <np.float64_t> ((self.pos[i] * self.nz) + self.ind[i])
            self.fcoords[self.index,i] = (c + 0.5) * dx
        self.index += 1

# Floating point widths; domain modifications are done later.
cdef class FWidthOcts(OctVisitor):
    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        # Note that this does not actually give the correct floating point
        # coordinates.  It gives them in some unit system where the domain is 1.0
        # in all directions, and assumes that they will be scaled later.
        if selected == 0: return
        cdef int i
        cdef np.float64_t dx
        dx = 1.0 / (self.nz << self.level)
        for i in range(3):
            self.fwidth[self.index,i] = dx
        self.index += 1

# Mark which domains are touched by a selector.
cdef class IdentifyOcts(OctVisitor):
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        # We assume that our domain has *already* been selected by, which means
        # we'll get all cells within the domain for a by-domain selector and all
        # cells within the domain *and* selector for the selector itself.
        if selected == 0: return
        self.domain_mask[o.domain - 1] = 1

# Assign domain indices to octs
cdef class AssignDomainInd(OctVisitor):
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        o.domain_ind = self.global_index
        self.index += 1

# From the file, fill in C order
cdef class FillFileIndicesO(OctVisitor):
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        # We fill these arrays, then inside the level filler we use these as
        # indices as we fill a second array from the self.
        if selected == 0: return
        self.levels[self.index] = self.level
        self.file_inds[self.index] = o.file_ind
        self.cell_inds[self.index] = self.oind()
        self.index +=1

# From the file, fill in F order
cdef class FillFileIndicesR(OctVisitor):
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        # We fill these arrays, then inside the level filler we use these as
        # indices as we fill a second array from the self.
        if selected == 0: return
        self.levels[self.index] = self.level
        self.file_inds[self.index] = o.file_ind
        self.cell_inds[self.index] = self.rind()
        self.index +=1

# Count octs by domain
cdef class CountByDomain(OctVisitor):
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        if selected == 0: return
        # NOTE: We do this for every *cell*.
        self.domain_counts[o.domain - 1] += 1

# Store the refinement mapping of the octree to be loaded later
cdef class StoreOctree(OctVisitor):
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        if o.children == NULL:
            # Not refined.
            res = 0
        else:
            res = 1
        self.ref_mask[self.index] = res
        self.index += 1

# Go from a refinement mapping to a new octree
cdef class LoadOctree(OctVisitor):
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        cdef int i, ii
        ii = cind(self.ind[0], self.ind[1], self.ind[2])
        if self.level > self.max_level:
            self.max_level = self.level
        if self.ref_mask[self.index] == 0:
            # We only want to do this once.  Otherwise we end up with way too many
            # nfinest for our tastes.
            if o.file_ind == -1:
                o.children = NULL
                o.file_ind = self.nfinest[0]
                o.domain = 1
                self.nfinest[0] += 1
        elif self.ref_mask[self.index] > 0:
            if self.ref_mask[self.index] != 1 and self.ref_mask[self.index] != 8:
                print("ARRAY CLUE: ", self.ref_mask[self.index], "UNKNOWN")
                raise RuntimeError
            if o.children == NULL:
                o.children = <Oct **> malloc(sizeof(Oct *) * 8)
                for i in range(8):
                    o.children[i] = NULL
            for i in range(8):
                o.children[ii + i] = &self.octs[self.nocts[0]]
                o.children[ii + i].domain_ind = self.nocts[0]
                o.children[ii + i].file_ind = -1
                o.children[ii + i].domain = -1
                o.children[ii + i].children = NULL
                self.nocts[0] += 1
        else:
            print("SOMETHING IS AMISS", self.index)
            raise RuntimeError
        self.index += 1

cdef class MortonIndexOcts(OctVisitor):
    @cython.boundscheck(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        if selected == 0: return
        cdef np.int64_t coord[3]
        cdef int i
        for i in range(3):
            coord[i] = (self.pos[i] * self.nz) + self.ind[i]
            if (coord[i] < 0):
                raise RuntimeError("Oct coordinate in dimension {} is ".format(i)+
                                   "negative. ({})".format(coord[i]))
        self.level_arr[self.index] = self.level
        self.morton_ind[self.index] = encode_morton_64bit(
                np.uint64(coord[0]),
                np.uint64(coord[1]),
                np.uint64(coord[2]))
        self.index += 1

# Store cell index
cdef class StoreIndex(OctVisitor):
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        if not selected: return
        self.cell_inds[o.domain_ind, self.ind[2], self.ind[1], self.ind[0]] = self.index

        self.index += 1

cdef class BaseNeighbourVisitor(OctVisitor):
    def __init__(self, OctreeContainer octree, int domain_id=-1, int n_ghost_zones=1):
        self.octree = octree
        self.neigh_ind[0] = 0
        self.neigh_ind[1] = 0
        self.neigh_ind[2] = 0
        self.n_ghost_zones = n_ghost_zones
        super(BaseNeighbourVisitor, self).__init__(octree, domain_id)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void set_neighbour_info(self, Oct *o, int ishift[3]):
        cdef int i
        cdef np.float64_t c, dx
        cdef np.int64_t ipos
        cdef np.float64_t fcoords[3]
        cdef Oct *neighbour
        cdef bint local_oct
        cdef bint other_oct
        dx = 1.0 / (self.nz << self.level)
        local_oct = True

        # Compute position of neighbouring cell
        for i in range(3):
            c = <np.float64_t> (self.pos[i] * self.nz)
            fcoords[i] = (c + 0.5 + ishift[i]) * dx / self.octree.nn[i]
            # Assuming periodicity
            if fcoords[i] < 0:
                fcoords[i] += 1
            elif fcoords[i] > 1:
                fcoords[i] -= 1
            local_oct &= (0 <= ishift[i] <= 1)
        other_oct = not local_oct

        # Use octree to find neighbour
        if other_oct:
            neighbour = self.octree.get(fcoords, &self.oi, max_level=self.level)
        else:
            neighbour = o
            self.oi.level = self.level
            for i in range(3):
                self.oi.ipos[i] = (self.pos[i] * self.nz) + ishift[i]

        # Extra step - compute cell position in neighbouring oct (and store in oi.ipos)
        if self.oi.level == self.level - 1:
            for i in range(3):
                ipos = (((self.pos[i] * self.nz) + ishift[i])) >> 1
                if (self.oi.ipos[i] << 1) == ipos:
                    self.oi.ipos[i] = 0
                else:
                    self.oi.ipos[i] = 1
        self.neighbour = neighbour

        # Index of neighbouring cell within its oct
        for i in range(3):
            self.neigh_ind[i] = <np.uint8_t>(ishift[i]) % 2

        self.other_oct = other_oct
        if other_oct:
            if self.neighbour != NULL:
                if self.oi.level == self.level - 1:
                    # Position within neighbouring oct is stored in oi.ipos
                    for i in range(3):
                        self.neigh_ind[i] = self.oi.ipos[i]
                elif self.oi.level != self.level:
                    print('This should not happen! %s %s' % (self.oi.level, self.level))
                    self.neighbour = NULL

# Store neighbouring cell index in current cell
cdef class NeighbourCellIndexVisitor(BaseNeighbourVisitor):
    # This piece of code is very much optimizable. Here are possible routes to achieve
    # much better performance:

    # - Work oct by oct, which would reduce the number of neighbor lookup
    #   from 4³=64 to 3³=27,
    # - Use faster neighbor lookup method(s). For now, all searches are started from
    #   the root mesh down to leaf nodes, but we could instead go up the tree from the
    #   central oct then down to find all neighbors (see e.g.
    #   https://geidav.wordpress.com/2017/12/02/advanced-octrees-4-finding-neighbor-nodes/).
    # -  Pre-compute the face-neighbors of all octs.

    # Note that for the last point, algorithms exist that generate the neighbors of all
    # octs in O(1) time (https://link.springer.com/article/10.1007/s13319-015-0060-9)
    # during the octree construction.
    # Another possible solution would be to keep a unordered hash map of all the octs
    # indexed by their (3-integers) position. With such structure, finding a neighbor
    # takes O(1) time. This could even come as a replacement of the current
    # pointer-based octree structure.

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        cdef int i, j, k
        cdef int ishift[3]
        cdef np.uint8_t neigh_cell_ind
        cdef np.int64_t neigh_domain_ind
        if selected == 0: return
        # Work at oct level
        if self.last == o.domain_ind: return

        self.last = o.domain_ind

        cdef int i0, i1
        i0 = -self.n_ghost_zones
        i1 = 2 + self.n_ghost_zones
        # Loop over cells in and directly around oct
        for i in range(i0, i1):
            ishift[0] = i
            for j in range(i0, i1):
                ishift[1] = j
                for k in range(i0, i1):
                    ishift[2] = k
                    self.set_neighbour_info(o, ishift)

                    if not self.other_oct:
                        neigh_domain_ind = o.domain_ind
                        neigh_cell_ind   = self.neighbour_rind()
                    elif self.neighbour != NULL:
                        neigh_domain_ind = self.neighbour.domain_ind
                        neigh_cell_ind   = self.neighbour_rind()
                    else:
                        neigh_domain_ind = -1
                        neigh_cell_ind   = 8

                    self.cell_inds[self.index]   = neigh_cell_ind
                    self.domain_inds[self.index] = neigh_domain_ind

                    self.index += 1

# Store file position + cell of neighbouring cell in current cell
cdef class NeighbourCellVisitor(BaseNeighbourVisitor):

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void visit(self, Oct* o, np.uint8_t selected):
        cdef int i, j, k
        cdef int ishift[3]
        cdef np.int64_t neigh_file_ind
        cdef np.uint8_t neigh_cell_ind
        cdef np.int32_t neigh_domain
        cdef np.uint8_t neigh_level
        if selected == 0: return
        # Work at oct level
        if self.last == o.domain_ind: return

        self.last = o.domain_ind

        cdef int i0, i1
        i0 = -self.n_ghost_zones
        i1 = 2 + self.n_ghost_zones
        # Loop over cells in and directly around oct
        for i in range(i0, i1):
            ishift[0] = i
            for j in range(i0, i1):
                ishift[1] = j
                for k in range(i0, i1):
                    ishift[2] = k
                    self.set_neighbour_info(o, ishift)

                    if not self.other_oct:
                        neigh_level    = self.level
                        neigh_domain   = o.domain
                        neigh_file_ind = o.file_ind
                        neigh_cell_ind = self.neighbour_rind()
                    elif self.neighbour != NULL:
                        neigh_level    = self.oi.level
                        neigh_domain   = self.neighbour.domain
                        neigh_file_ind = self.neighbour.file_ind
                        neigh_cell_ind = self.neighbour_rind()
                    else:
                        neigh_level    = 255
                        neigh_domain   = -1
                        neigh_file_ind = -1
                        neigh_cell_ind = 8

                    self.levels[self.index]    = neigh_level
                    self.file_inds[self.index] = neigh_file_ind
                    self.cell_inds[self.index] = neigh_cell_ind
                    self.domains[self.index]   = neigh_domain

                    self.index += 1



import inspect
from collections import Counter
from typing import TypeVar

from more_itertools import flatten

from yt.data_objects.static_output import Dataset
from yt.utilities.object_registries import output_type_registry

T = TypeVar("T")


def find_lowest_subclasses(candidates: list[type[T]]) -> list[type[T]]:
    """
    This function takes a list of classes, and returns only the ones that are
    are not super classes of any others in the list. i.e. the ones that are at
    the bottom of the specified mro of classes.

    Parameters
    ----------
    candidates : Iterable
        An iterable object that is a collection of classes to find the lowest
        subclass of.

    Returns
    -------
    result : list
        A list of classes which are not super classes for any others in
        candidates.
    """
    count = Counter(flatten(inspect.getmro(c) for c in candidates))
    return [x for x in candidates if count[x] == 1]


def get_classes_with_missing_requirements() -> dict[type[Dataset], list[str]]:
    # We need a function here rather than an global constant registry because:
    # - computation should be delayed until needed so that the result is independent of import order
    # - tests might (temporarily) mutate output_type_registry
    return {
        cls: missing
        for cls in sorted(output_type_registry.values(), key=lambda c: c.__name__)
        if (missing := cls._missing_load_requirements())
    }



import os
import shutil
import tempfile
import unittest

import numpy as np
from nose.tools import assert_raises
from numpy.testing import assert_array_equal, assert_equal

from yt.data_objects.data_containers import YTDataContainer
from yt.data_objects.particle_filters import particle_filter
from yt.testing import (
    fake_amr_ds,
    fake_particle_ds,
    fake_random_ds,
    requires_module,
)
from yt.utilities.exceptions import YTException, YTFieldNotFound


class TestDataContainers(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tmpdir = tempfile.mkdtemp()
        cls.curdir = os.getcwd()
        os.chdir(cls.tmpdir)

    @classmethod
    def tearDownClass(cls):
        os.chdir(cls.curdir)
        shutil.rmtree(cls.tmpdir)

    def test_yt_data_container(self):
        # Test if ds could be None
        with assert_raises(RuntimeError) as err:
            YTDataContainer(None, None)
        desired = (
            "Error: ds must be set either through class"
            " type or parameter to the constructor"
        )
        assert_equal(str(err.exception), desired)

        # Test if field_data key exists
        ds = fake_random_ds(5)
        proj = ds.proj(("gas", "density"), 0, data_source=ds.all_data())
        assert_equal("px" in proj.keys(), True)
        assert_equal("pz" in proj.keys(), False)

        # Delete the key and check if exits
        del proj["px"]
        assert_equal("px" in proj.keys(), False)
        del proj["gas", "density"]
        assert_equal("density" in proj.keys(), False)

        # Delete a non-existent field
        with assert_raises(YTFieldNotFound) as ex:
            del proj["p_mass"]
        desired = "Could not find field 'p_mass' in UniformGridData."
        assert_equal(str(ex.exception), desired)

    def test_write_out(self):
        filename = "sphere.txt"
        ds = fake_random_ds(16, particles=10)
        sp = ds.sphere(ds.domain_center, 0.25)

        sp.write_out(filename, fields=[("gas", "cell_volume")])

        with open(filename) as file:
            file_row_1 = file.readline()
            file_row_2 = file.readline()
            file_row_2 = np.array(file_row_2.split("\t"), dtype=np.float64)
        sorted_keys = sorted(sp.field_data.keys())
        keys = [str(k) for k in sorted_keys]
        keys = "\t".join(["#"] + keys + ["\n"])
        data = [sp.field_data[k][0] for k in sorted_keys]

        assert_equal(keys, file_row_1)
        assert_array_equal(data, file_row_2)

    def test_invalid_write_out(self):
        filename = "sphere.txt"
        ds = fake_random_ds(16, particles=10)
        sp = ds.sphere(ds.domain_center, 0.25)

        with assert_raises(YTException):
            sp.write_out(filename, fields=[("all", "particle_ones")])

    @requires_module("pandas")
    def test_to_dataframe(self):
        fields = [("gas", "density"), ("gas", "velocity_z")]
        ds = fake_random_ds(6)
        dd = ds.all_data()
        df = dd.to_dataframe(fields)
        assert_array_equal(dd[fields[0]], df[fields[0][1]])
        assert_array_equal(dd[fields[1]], df[fields[1][1]])

    @requires_module("astropy")
    def test_to_astropy_table(self):
        from yt.units.yt_array import YTArray

        fields = [("gas", "density"), ("gas", "velocity_z")]
        ds = fake_random_ds(6)
        dd = ds.all_data()
        at1 = dd.to_astropy_table(fields)
        assert_array_equal(dd[fields[0]].d, at1[fields[0][1]].value)
        assert_array_equal(dd[fields[1]].d, at1[fields[1][1]].value)
        assert dd[fields[0]].units == YTArray.from_astropy(at1[fields[0][1]]).units
        assert dd[fields[1]].units == YTArray.from_astropy(at1[fields[1][1]]).units

    def test_std(self):
        ds = fake_random_ds(3)
        ds.all_data().std(("gas", "density"), weight=("gas", "velocity_z"))

    def test_to_frb(self):
        # Test cylindrical geometry
        fields = ["density", "cell_mass"]
        units = ["g/cm**3", "g"]
        ds = fake_amr_ds(
            fields=fields, units=units, geometry="cylindrical", particles=16**3
        )
        dd = ds.all_data()
        proj = ds.proj(
            ("gas", "density"),
            weight_field=("gas", "cell_mass"),
            axis=1,
            data_source=dd,
        )
        frb = proj.to_frb((1.0, "unitary"), 64)
        assert_equal(frb.radius, (1.0, "unitary"))
        assert_equal(frb.buff_size, 64)

    def test_extract_isocontours(self):
        # Test isocontour properties for AMRGridData
        fields = ["density", "cell_mass"]
        units = ["g/cm**3", "g"]
        ds = fake_amr_ds(fields=fields, units=units, particles=16**3)
        dd = ds.all_data()
        q = dd.quantities["WeightedAverageQuantity"]
        rho = q(("gas", "density"), weight=("gas", "cell_mass"))
        dd.extract_isocontours(("gas", "density"), rho, "triangles.obj", True)
        dd.calculate_isocontour_flux(
            ("gas", "density"),
            rho,
            ("index", "x"),
            ("index", "y"),
            ("index", "z"),
            ("index", "dx"),
        )

        # Test error in case of ParticleData
        ds = fake_particle_ds()
        dd = ds.all_data()
        q = dd.quantities["WeightedAverageQuantity"]
        rho = q(("all", "particle_velocity_x"), weight=("all", "particle_mass"))
        with assert_raises(NotImplementedError):
            dd.extract_isocontours("density", rho, sample_values="x")

    def test_derived_field(self):
        # Test that derived field on filtered particles do not require
        # their parent field to be created
        ds = fake_particle_ds()
        dd = ds.all_data()
        dd.set_field_parameter("axis", 0)

        @particle_filter(requires=["particle_mass"], filtered_type="io")
        def massive(pfilter, data):
            return data[pfilter.filtered_type, "particle_mass"].to("code_mass") > 0.5

        ds.add_particle_filter("massive")

        def fun(field, data):
            return data[field.name[0], "particle_mass"]

        # Add the field to the massive particles
        ds.add_field(
            ("massive", "test"),
            function=fun,
            sampling_type="particle",
            units="code_mass",
        )

        expected_size = (dd["io", "particle_mass"].to("code_mass") > 0.5).sum()

        fields_to_test = [f for f in ds.derived_field_list if f[0] == "massive"]

        def test_this(fname):
            data = dd[fname]
            assert_equal(data.shape[0], expected_size)

        for fname in fields_to_test:
            test_this(fname)



from yt._maintenance.deprecation import issue_deprecation_warning, warnings


def boxlib_deprecation():
    with warnings.catch_warnings():
        warnings.simplefilter("always")
        issue_deprecation_warning(
            "The historic 'boxlib' frontend is \n"
            "deprecated as it has been renamed 'amrex'. "
            "Existing and future work should instead reference the 'amrex' frontend.",
            stacklevel=4,
            since="4.4.0",
        )



import abc
import itertools
import sys
import uuid
from collections import defaultdict
from contextlib import contextmanager

import numpy as np
from more_itertools import always_iterable
from unyt import unyt_array
from unyt.exceptions import UnitConversionError, UnitParseError

import yt.geometry
from yt.data_objects.data_containers import YTDataContainer
from yt.data_objects.derived_quantities import DerivedQuantityCollection
from yt.data_objects.field_data import YTFieldData
from yt.fields.field_exceptions import NeedsGridType
from yt.funcs import fix_axis, is_sequence, iter_fields, validate_width_tuple
from yt.geometry.api import Geometry
from yt.geometry.selection_routines import compose_selector
from yt.units import YTArray
from yt.utilities.exceptions import (
    GenerationInProgress,
    YTBooleanObjectError,
    YTBooleanObjectsWrongDataset,
    YTDataSelectorNotImplemented,
    YTDimensionalityError,
    YTFieldUnitError,
    YTFieldUnitParseError,
)
from yt.utilities.lib.marching_cubes import march_cubes_grid, march_cubes_grid_flux
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.parallel_tools.parallel_analysis_interface import (
    ParallelAnalysisInterface,
)

if sys.version_info >= (3, 11):
    from typing import assert_never
else:
    from typing_extensions import assert_never


class YTSelectionContainer(YTDataContainer, ParallelAnalysisInterface, abc.ABC):
    _locked = False
    _sort_by = None
    _selector = None
    _current_chunk = None
    _data_source = None
    _dimensionality: int
    _max_level = None
    _min_level = None
    _derived_quantity_chunking = "io"

    def __init__(self, ds, field_parameters, data_source=None):
        ParallelAnalysisInterface.__init__(self)
        super().__init__(ds, field_parameters)
        self._data_source = data_source
        if data_source is not None:
            if data_source.ds != self.ds:
                raise RuntimeError(
                    "Attempted to construct a DataContainer with a data_source "
                    "from a different Dataset",
                    ds,
                    data_source.ds,
                )
            if data_source._dimensionality < self._dimensionality:
                raise RuntimeError(
                    "Attempted to construct a DataContainer with a data_source "
                    "of lower dimensionality (%u vs %u)"
                    % (data_source._dimensionality, self._dimensionality)
                )
            self.field_parameters.update(data_source.field_parameters)
        self.quantities = DerivedQuantityCollection(self)

    @property
    def selector(self):
        if self._selector is not None:
            return self._selector
        s_module = getattr(self, "_selector_module", yt.geometry.selection_routines)
        sclass = getattr(s_module, f"{self._type_name}_selector", None)
        if sclass is None:
            raise YTDataSelectorNotImplemented(self._type_name)

        if self._data_source is not None:
            self._selector = compose_selector(
                self, self._data_source.selector, sclass(self)
            )
        else:
            self._selector = sclass(self)
        return self._selector

    def chunks(self, fields, chunking_style, **kwargs):
        # This is an iterator that will yield the necessary chunks.
        self.get_data()  # Ensure we have built ourselves
        if fields is None:
            fields = []
        # chunk_ind can be supplied in the keyword arguments.  If it's a
        # scalar, that'll be the only chunk that gets returned; if it's a list,
        # those are the ones that will be.
        chunk_ind = kwargs.pop("chunk_ind", None)
        if chunk_ind is not None:
            chunk_ind = list(always_iterable(chunk_ind))
        for ci, chunk in enumerate(self.index._chunk(self, chunking_style, **kwargs)):
            if chunk_ind is not None and ci not in chunk_ind:
                continue
            with self._chunked_read(chunk):
                self.get_data(fields)
                # NOTE: we yield before releasing the context
                yield self

    def _identify_dependencies(self, fields_to_get, spatial=False):
        inspected = 0
        fields_to_get = fields_to_get[:]
        for field in itertools.cycle(fields_to_get):
            if inspected >= len(fields_to_get):
                break
            inspected += 1
            fi = self.ds._get_field_info(field)
            fd = self.ds.field_dependencies.get(
                field, None
            ) or self.ds.field_dependencies.get(field[1], None)
            # This is long overdue.  Any time we *can't* find a field
            # dependency -- for instance, if the derived field has been added
            # after dataset instantiation -- let's just try to
            # recalculate it.
            if fd is None:
                try:
                    fd = fi.get_dependencies(ds=self.ds)
                    self.ds.field_dependencies[field] = fd
                except Exception:
                    continue
            requested = self._determine_fields(list(set(fd.requested)))
            deps = [d for d in requested if d not in fields_to_get]
            fields_to_get += deps
        return sorted(fields_to_get)

    def get_data(self, fields=None):
        if self._current_chunk is None:
            self.index._identify_base_chunk(self)
        if fields is None:
            return
        nfields = []
        apply_fields = defaultdict(list)
        for field in self._determine_fields(fields):
            # We need to create the field on the raw particle types
            # for particles types (when the field is not directly
            # defined for the derived particle type only)
            finfo = self.ds.field_info[field]

            if (
                field[0] in self.ds.filtered_particle_types
                and finfo._inherited_particle_filter
            ):
                f = self.ds.known_filters[field[0]]
                apply_fields[field[0]].append((f.filtered_type, field[1]))
            else:
                nfields.append(field)
        for filter_type in apply_fields:
            f = self.ds.known_filters[filter_type]
            with f.apply(self):
                self.get_data(apply_fields[filter_type])
        fields = nfields
        if len(fields) == 0:
            return
        # Now we collect all our fields
        # Here is where we need to perform a validation step, so that if we
        # have a field requested that we actually *can't* yet get, we put it
        # off until the end.  This prevents double-reading fields that will
        # need to be used in spatial fields later on.
        fields_to_get = []
        # This will be pre-populated with spatial fields
        fields_to_generate = []
        for field in self._determine_fields(fields):
            if field in self.field_data:
                continue
            finfo = self.ds._get_field_info(field)
            try:
                finfo.check_available(self)
            except NeedsGridType:
                fields_to_generate.append(field)
                continue
            fields_to_get.append(field)
        if len(fields_to_get) == 0 and len(fields_to_generate) == 0:
            return
        elif self._locked:
            raise GenerationInProgress(fields)
        # Track which ones we want in the end
        ofields = set(list(self.field_data.keys()) + fields_to_get + fields_to_generate)
        # At this point, we want to figure out *all* our dependencies.
        fields_to_get = self._identify_dependencies(fields_to_get, self._spatial)
        # We now split up into readers for the types of fields
        fluids, particles = [], []
        finfos = {}
        for field_key in fields_to_get:
            finfo = self.ds._get_field_info(field_key)
            finfos[field_key] = finfo
            if finfo.sampling_type == "particle":
                particles.append(field_key)
            elif field_key not in fluids:
                fluids.append(field_key)
        # The _read method will figure out which fields it needs to get from
        # disk, and return a dict of those fields along with the fields that
        # need to be generated.
        read_fluids, gen_fluids = self.index._read_fluid_fields(
            fluids, self, self._current_chunk
        )
        for f, v in read_fluids.items():
            self.field_data[f] = self.ds.arr(v, units=finfos[f].units)
            self.field_data[f].convert_to_units(finfos[f].output_units)

        read_particles, gen_particles = self.index._read_particle_fields(
            particles, self, self._current_chunk
        )

        for f, v in read_particles.items():
            self.field_data[f] = self.ds.arr(v, units=finfos[f].units)
            self.field_data[f].convert_to_units(finfos[f].output_units)

        fields_to_generate += gen_fluids + gen_particles
        self._generate_fields(fields_to_generate)
        for field in list(self.field_data.keys()):
            if field not in ofields:
                self.field_data.pop(field)

    def _generate_fields(self, fields_to_generate):
        index = 0

        def dimensions_compare_equal(a, b, /) -> bool:
            if a == b:
                return True
            try:
                if (a == 1 and b.is_dimensionless) or (a.is_dimensionless and b == 1):
                    return True
            except AttributeError:
                return False
            return False

        with self._field_lock():
            # At this point, we assume that any fields that are necessary to
            # *generate* a field are in fact already available to us.  Note
            # that we do not make any assumption about whether or not the
            # fields have a spatial requirement.  This will be checked inside
            # _generate_field, at which point additional dependencies may
            # actually be noted.
            while any(f not in self.field_data for f in fields_to_generate):
                field = fields_to_generate[index % len(fields_to_generate)]
                index += 1
                if field in self.field_data:
                    continue
                fi = self.ds._get_field_info(field)
                try:
                    fd = self._generate_field(field)
                    if hasattr(fd, "units"):
                        fd.units.registry = self.ds.unit_registry
                    if fd is None:
                        raise RuntimeError
                    if fi.units is None:
                        # first time calling a field with units='auto', so we
                        # infer the units from the units of the data we get back
                        # from the field function and use these units for future
                        # field accesses
                        units = getattr(fd, "units", "")
                        if units == "":
                            sunits = ""
                            dimensions = 1
                        else:
                            sunits = str(
                                units.get_base_equivalent(self.ds.unit_system.name)
                            )
                            dimensions = units.dimensions

                        if fi.dimensions is None:
                            mylog.warning(
                                "Field %s was added without specifying units or dimensions, "
                                "auto setting units to %r",
                                fi.name,
                                sunits or "dimensionless",
                            )
                        elif not dimensions_compare_equal(fi.dimensions, dimensions):
                            raise YTDimensionalityError(fi.dimensions, dimensions)
                        fi.units = sunits
                        fi.dimensions = dimensions
                        self.field_data[field] = self.ds.arr(fd, units)
                    if fi.output_units is None:
                        fi.output_units = fi.units

                    try:
                        fd.convert_to_units(fi.units)
                    except AttributeError:
                        # If the field returns an ndarray, coerce to a
                        # dimensionless YTArray and verify that field is
                        # supposed to be unitless
                        fd = self.ds.arr(fd, "")
                        if fi.units != "":
                            raise YTFieldUnitError(fi, fd.units) from None
                    except UnitConversionError as e:
                        raise YTFieldUnitError(fi, fd.units) from e
                    except UnitParseError as e:
                        raise YTFieldUnitParseError(fi) from e
                    self.field_data[field] = fd
                except GenerationInProgress as gip:
                    for f in gip.fields:
                        if f not in fields_to_generate:
                            fields_to_generate.append(f)

    def __or__(self, other):
        if not isinstance(other, YTSelectionContainer):
            raise YTBooleanObjectError(other)
        if self.ds is not other.ds:
            raise YTBooleanObjectsWrongDataset()
        # Should maybe do something with field parameters here
        from yt.data_objects.selection_objects.boolean_operations import (
            YTBooleanContainer,
        )

        return YTBooleanContainer("OR", self, other, ds=self.ds)

    def __invert__(self):
        # ~obj
        asel = yt.geometry.selection_routines.AlwaysSelector(self.ds)
        from yt.data_objects.selection_objects.boolean_operations import (
            YTBooleanContainer,
        )

        return YTBooleanContainer("NOT", self, asel, ds=self.ds)

    def __xor__(self, other):
        if not isinstance(other, YTSelectionContainer):
            raise YTBooleanObjectError(other)
        if self.ds is not other.ds:
            raise YTBooleanObjectsWrongDataset()
        from yt.data_objects.selection_objects.boolean_operations import (
            YTBooleanContainer,
        )

        return YTBooleanContainer("XOR", self, other, ds=self.ds)

    def __and__(self, other):
        if not isinstance(other, YTSelectionContainer):
            raise YTBooleanObjectError(other)
        if self.ds is not other.ds:
            raise YTBooleanObjectsWrongDataset()
        from yt.data_objects.selection_objects.boolean_operations import (
            YTBooleanContainer,
        )

        return YTBooleanContainer("AND", self, other, ds=self.ds)

    def __add__(self, other):
        return self.__or__(other)

    def __sub__(self, other):
        if not isinstance(other, YTSelectionContainer):
            raise YTBooleanObjectError(other)
        if self.ds is not other.ds:
            raise YTBooleanObjectsWrongDataset()
        from yt.data_objects.selection_objects.boolean_operations import (
            YTBooleanContainer,
        )

        return YTBooleanContainer("NEG", self, other, ds=self.ds)

    @contextmanager
    def _field_lock(self):
        self._locked = True
        yield
        self._locked = False

    @contextmanager
    def _ds_hold(self, new_ds):
        """
        This contextmanager is used to take a data object and preserve its
        attributes but allow the dataset that underlies it to be swapped out.
        This is typically only used internally, and differences in unit systems
        may present interesting possibilities.
        """
        old_ds = self.ds
        old_index = self._index
        self.ds = new_ds
        self._index = new_ds.index
        old_chunk_info = self._chunk_info
        old_chunk = self._current_chunk
        old_size = self.size
        self._chunk_info = None
        self._current_chunk = None
        self.size = None
        self._index._identify_base_chunk(self)
        with self._chunked_read(None):
            yield
        self._index = old_index
        self.ds = old_ds
        self._chunk_info = old_chunk_info
        self._current_chunk = old_chunk
        self.size = old_size

    @contextmanager
    def _chunked_read(self, chunk):
        # There are several items that need to be swapped out
        # field_data, size, shape
        obj_field_data = []
        if hasattr(chunk, "objs"):
            for obj in chunk.objs:
                obj_field_data.append(obj.field_data)
                obj.field_data = YTFieldData()
        old_field_data, self.field_data = self.field_data, YTFieldData()
        old_chunk, self._current_chunk = self._current_chunk, chunk
        old_locked, self._locked = self._locked, False
        yield
        self.field_data = old_field_data
        self._current_chunk = old_chunk
        self._locked = old_locked
        if hasattr(chunk, "objs"):
            for obj in chunk.objs:
                obj.field_data = obj_field_data.pop(0)

    @contextmanager
    def _activate_cache(self):
        cache = self._field_cache or {}
        old_fields = {}
        for field in (f for f in cache if f in self.field_data):
            old_fields[field] = self.field_data[field]
        self.field_data.update(cache)
        yield
        for field in cache:
            self.field_data.pop(field)
            if field in old_fields:
                self.field_data[field] = old_fields.pop(field)
        self._field_cache = None

    def _initialize_cache(self, cache):
        # Wipe out what came before
        self._field_cache = {}
        self._field_cache.update(cache)

    @property
    def icoords(self):
        if self._current_chunk is None:
            self.index._identify_base_chunk(self)
        return self._current_chunk.icoords

    @property
    def fcoords(self):
        if self._current_chunk is None:
            self.index._identify_base_chunk(self)
        return self._current_chunk.fcoords

    @property
    def ires(self):
        if self._current_chunk is None:
            self.index._identify_base_chunk(self)
        return self._current_chunk.ires

    @property
    def fwidth(self):
        if self._current_chunk is None:
            self.index._identify_base_chunk(self)
        return self._current_chunk.fwidth

    @property
    def fcoords_vertex(self):
        if self._current_chunk is None:
            self.index._identify_base_chunk(self)
        return self._current_chunk.fcoords_vertex

    @property
    def max_level(self):
        if self._max_level is None:
            try:
                return self.ds.max_level
            except AttributeError:
                return None
        return self._max_level

    @max_level.setter
    def max_level(self, value):
        if self._selector is not None:
            del self._selector
            self._selector = None
        self._current_chunk = None
        self.size = None
        self.shape = None
        self.field_data.clear()
        self._max_level = value

    @property
    def min_level(self):
        if self._min_level is None:
            try:
                return 0
            except AttributeError:
                return None
        return self._min_level

    @min_level.setter
    def min_level(self, value):
        if self._selector is not None:
            del self._selector
            self._selector = None
        self.field_data.clear()
        self.size = None
        self.shape = None
        self._current_chunk = None
        self._min_level = value


class YTSelectionContainer0D(YTSelectionContainer):
    _spatial = False
    _dimensionality = 0

    def __init__(self, ds, field_parameters=None, data_source=None):
        super().__init__(ds, field_parameters, data_source)


class YTSelectionContainer1D(YTSelectionContainer):
    _spatial = False
    _dimensionality = 1

    def __init__(self, ds, field_parameters=None, data_source=None):
        super().__init__(ds, field_parameters, data_source)
        self._grids = None
        self._sortkey = None
        self._sorted = {}


class YTSelectionContainer2D(YTSelectionContainer):
    _key_fields = ["px", "py", "pdx", "pdy"]
    _dimensionality = 2
    """
    Prepares the YTSelectionContainer2D, normal to *axis*.  If *axis* is 4, we are not
    aligned with any axis.
    """
    _spatial = False

    def __init__(self, axis, ds, field_parameters=None, data_source=None):
        super().__init__(ds, field_parameters, data_source)
        # We need the ds, which will exist by now, for fix_axis.
        self.axis = fix_axis(axis, self.ds)
        self.set_field_parameter("axis", axis)

    def _convert_field_name(self, field):
        return field

    def _get_pw(self, fields, center, width, origin, plot_type):
        from yt.visualization.fixed_resolution import FixedResolutionBuffer as frb
        from yt.visualization.plot_window import PWViewerMPL, get_window_parameters

        axis = self.axis
        skip = self._key_fields
        skip += list(set(frb._exclude_fields).difference(set(self._key_fields)))
        self.fields = [k for k in self.field_data if k not in skip]
        if fields is not None:
            self.fields = list(iter_fields(fields)) + self.fields
        if len(self.fields) == 0:
            raise ValueError("No fields found to plot in get_pw")
        (bounds, center, display_center) = get_window_parameters(
            axis, center, width, self.ds
        )
        pw = PWViewerMPL(
            self,
            bounds,
            fields=self.fields,
            origin=origin,
            frb_generator=frb,
            plot_type=plot_type,
            geometry=self.ds.geometry,
        )
        pw._setup_plots()
        return pw

    def to_frb(self, width, resolution, center=None, height=None, periodic=False):
        r"""This function returns a FixedResolutionBuffer generated from this
        object.

        A FixedResolutionBuffer is an object that accepts a variable-resolution
        2D object and transforms it into an NxM bitmap that can be plotted,
        examined or processed.  This is a convenience function to return an FRB
        directly from an existing 2D data object.

        Parameters
        ----------
        width : width specifier
            This can either be a floating point value, in the native domain
            units of the simulation, or a tuple of the (value, unit) style.
            This will be the width of the FRB.
        height : height specifier
            This will be the physical height of the FRB, by default it is equal
            to width.  Note that this will not make any corrections to
            resolution for the aspect ratio.
        resolution : int or tuple of ints
            The number of pixels on a side of the final FRB.  If iterable, this
            will be the width then the height.
        center : array-like of floats, optional
            The center of the FRB.  If not specified, defaults to the center of
            the current object.
        periodic : bool
            Should the returned Fixed Resolution Buffer be periodic?  (default:
            False).

        Returns
        -------
        frb : :class:`~yt.visualization.fixed_resolution.FixedResolutionBuffer`
            A fixed resolution buffer, which can be queried for fields.

        Examples
        --------

        >>> proj = ds.proj(("gas", "density"), 0)
        >>> frb = proj.to_frb((100.0, "kpc"), 1024)
        >>> write_image(np.log10(frb["gas", "density"]), "density_100kpc.png")
        """

        if (self.ds.geometry is Geometry.CYLINDRICAL and self.axis == 1) or (
            self.ds.geometry is Geometry.POLAR and self.axis == 2
        ):
            if center is not None and center != (0.0, 0.0):
                raise NotImplementedError(
                    "Currently we only support images centered at R=0. "
                    + "We plan to generalize this in the near future"
                )
            from yt.visualization.fixed_resolution import (
                CylindricalFixedResolutionBuffer,
            )

            validate_width_tuple(width)
            if is_sequence(resolution):
                resolution = max(resolution)
            frb = CylindricalFixedResolutionBuffer(self, width, resolution)
            return frb

        if center is None:
            center = self.center
            if center is None:
                center = (self.ds.domain_right_edge + self.ds.domain_left_edge) / 2.0
        elif is_sequence(center) and not isinstance(center, YTArray):
            center = self.ds.arr(center, "code_length")
        if is_sequence(width):
            w, u = width
            if isinstance(w, tuple) and isinstance(u, tuple):
                height = u
                w, u = w
            width = self.ds.quan(w, units=u)
        elif not isinstance(width, YTArray):
            width = self.ds.quan(width, "code_length")
        if height is None:
            height = width
        elif is_sequence(height):
            h, u = height
            height = self.ds.quan(h, units=u)
        elif not isinstance(height, YTArray):
            height = self.ds.quan(height, "code_length")
        if not is_sequence(resolution):
            resolution = (resolution, resolution)
        from yt.visualization.fixed_resolution import FixedResolutionBuffer

        xax = self.ds.coordinates.x_axis[self.axis]
        yax = self.ds.coordinates.y_axis[self.axis]
        bounds = (
            center[xax] - width * 0.5,
            center[xax] + width * 0.5,
            center[yax] - height * 0.5,
            center[yax] + height * 0.5,
        )
        frb = FixedResolutionBuffer(self, bounds, resolution, periodic=periodic)
        return frb


class YTSelectionContainer3D(YTSelectionContainer):
    """
    Returns an instance of YTSelectionContainer3D, or prepares one.  Usually only
    used as a base class.  Note that *center* is supplied, but only used
    for fields and quantities that require it.
    """

    _key_fields = ["x", "y", "z", "dx", "dy", "dz"]
    _spatial = False
    _num_ghost_zones = 0
    _dimensionality = 3

    def __init__(self, center, ds, field_parameters=None, data_source=None):
        super().__init__(ds, field_parameters, data_source)
        self._set_center(center)
        self.coords = None
        self._grids = None

    def cut_region(self, field_cuts, field_parameters=None, locals=None):
        """
        Return a YTCutRegion, where the a cell is identified as being inside
        the cut region based on the value of one or more fields.  Note that in
        previous versions of yt the name 'grid' was used to represent the data
        object used to construct the field cut, as of yt 3.0, this has been
        changed to 'obj'.

        Parameters
        ----------
        field_cuts : list of strings
           A list of conditionals that will be evaluated. In the namespace
           available, these conditionals will have access to 'obj' which is a
           data object of unknown shape, and they must generate a boolean array.
           For instance, conditionals = ["obj['gas', 'temperature'] < 1e3"]
        field_parameters : dictionary
           A dictionary of field parameters to be used when applying the field
           cuts.
        locals : dictionary
            A dictionary of local variables to use when defining the cut region.

        Examples
        --------
        To find the total mass of hot gas with temperature greater than 10^6 K
        in your volume:

        >>> ds = yt.load("RedshiftOutput0005")
        >>> ad = ds.all_data()
        >>> cr = ad.cut_region(["obj['gas', 'temperature'] > 1e6"])
        >>> print(cr.quantities.total_quantity(("gas", "cell_mass")).in_units("Msun"))
        """
        if locals is None:
            locals = {}
        cr = self.ds.cut_region(
            self, field_cuts, field_parameters=field_parameters, locals=locals
        )
        return cr

    def _build_operator_cut(self, operation, field, value, units=None):
        """
        Given an operation (>, >=, etc.), a field and a value,
        return the cut_region implementing it.

        This is only meant to be used internally.

        Examples
        --------
        >>> ds._build_operator_cut(">", ("gas", "density"), 1e-24)
        ... # is equivalent to
        ... ds.cut_region(['obj["gas", "density"] > 1e-24'])
        """
        ftype, fname = self._determine_fields(field)[0]
        if units is None:
            field_cuts = f'obj["{ftype}", "{fname}"] {operation} {value}'
        else:
            field_cuts = (
                f'obj["{ftype}", "{fname}"].in_units("{units}") {operation} {value}'
            )
        return self.cut_region(field_cuts)

    def _build_function_cut(self, function, field, units=None, **kwargs):
        """
        Given a function (np.abs, np.all) and a field,
        return the cut_region implementing it.

        This is only meant to be used internally.

        Examples
        --------
        >>> ds._build_function_cut("np.isnan", ("gas", "density"), locals={"np": np})
        ... # is equivalent to
        ... ds.cut_region(['np.isnan(obj["gas", "density"])'], locals={"np": np})
        """
        ftype, fname = self._determine_fields(field)[0]
        if units is None:
            field_cuts = f'{function}(obj["{ftype}", "{fname}"])'
        else:
            field_cuts = f'{function}(obj["{ftype}", "{fname}"].in_units("{units}"))'
        return self.cut_region(field_cuts, **kwargs)

    def exclude_above(self, field, value, units=None):
        """
        This function will return a YTCutRegion where all of the regions
        whose field is above a given value are masked.

        Parameters
        ----------
        field : string
            The field in which the conditional will be applied.
        value : float
            The minimum value that will not be masked in the output
            YTCutRegion.
        units : string or None
            The units of the value threshold. None will use the default units
            given in the field.

        Returns
        -------
        cut_region : YTCutRegion
            The YTCutRegion with the field above the given value masked.

        Examples
        --------
        To find the total mass of hot gas with temperature colder than 10^6 K
        in your volume:

        >>> ds = yt.load("RedshiftOutput0005")
        >>> ad = ds.all_data()
        >>> cr = ad.exclude_above(("gas", "temperature"), 1e6)
        >>> print(cr.quantities.total_quantity(("gas", "cell_mass")).in_units("Msun"))

        """
        return self._build_operator_cut("<=", field, value, units)

    def include_above(self, field, value, units=None):
        """
        This function will return a YTCutRegion where only the regions
        whose field is above a given value are included.

        Parameters
        ----------
        field : string
            The field in which the conditional will be applied.
        value : float
            The minimum value that will not be masked in the output
            YTCutRegion.
        units : string or None
            The units of the value threshold. None will use the default units
            given in the field.

        Returns
        -------
        cut_region : YTCutRegion
            The YTCutRegion with the field above the given value masked.

        Examples
        --------
        To find the total mass of hot gas with temperature warmer than 10^6 K
        in your volume:

        >>> ds = yt.load("RedshiftOutput0005")
        >>> ad = ds.all_data()
        >>> cr = ad.include_above(("gas", "temperature"), 1e6)
        >>> print(cr.quantities.total_quantity(("gas", "cell_mass")).in_units("Msun"))
        """

        return self._build_operator_cut(">", field, value, units)

    def exclude_equal(self, field, value, units=None):
        """
        This function will return a YTCutRegion where all of the regions
        whose field are equal to given value are masked.

        Parameters
        ----------
        field : string
            The field in which the conditional will be applied.
        value : float
            The minimum value that will not be masked in the output
            YTCutRegion.
        units : string or None
            The units of the value threshold. None will use the default units
            given in the field.

        Returns
        -------
        cut_region : YTCutRegion
            The YTCutRegion with the field equal to the given value masked.

        Examples
        --------
        >>> ds = yt.load("RedshiftOutput0005")
        >>> ad = ds.all_data()
        >>> cr = ad.exclude_equal(("gas", "temperature"), 1e6)
        >>> print(cr.quantities.total_quantity(("gas", "cell_mass")).in_units("Msun"))
        """
        return self._build_operator_cut("!=", field, value, units)

    def include_equal(self, field, value, units=None):
        """
        This function will return a YTCutRegion where only the regions
        whose field are equal to given value are included.

        Parameters
        ----------
        field : string
            The field in which the conditional will be applied.
        value : float
            The minimum value that will not be masked in the output
            YTCutRegion.
        units : string or None
            The units of the value threshold. None will use the default units
            given in the field.

        Returns
        -------
        cut_region : YTCutRegion
            The YTCutRegion with the field equal to the given value included.

        Examples
        --------
        >>> ds = yt.load("RedshiftOutput0005")
        >>> ad = ds.all_data()
        >>> cr = ad.include_equal(("gas", "temperature"), 1e6)
        >>> print(cr.quantities.total_quantity(("gas", "cell_mass")).in_units("Msun"))
        """
        return self._build_operator_cut("==", field, value, units)

    def exclude_inside(self, field, min_value, max_value, units=None):
        """
        This function will return a YTCutRegion where all of the regions
        whose field are inside the interval from min_value to max_value.

        Parameters
        ----------
        field : string
            The field in which the conditional will be applied.
        min_value : float
            The minimum value inside the interval to be excluded.
        max_value : float
            The maximum value inside the interval to be excluded.
        units : string or None
            The units of the value threshold. None will use the default units
            given in the field.

        Returns
        -------
        cut_region : YTCutRegion
            The YTCutRegion with the field inside the given interval excluded.

        Examples
        --------
        >>> ds = yt.load("RedshiftOutput0005")
        >>> ad = ds.all_data()
        >>> cr = ad.exclude_inside(("gas", "temperature"), 1e5, 1e6)
        >>> print(cr.quantities.total_quantity(("gas", "cell_mass")).in_units("Msun"))
        """
        ftype, fname = self._determine_fields(field)[0]
        if units is None:
            field_cuts = (
                f'(obj["{ftype}", "{fname}"] <= {min_value}) | '
                f'(obj["{ftype}", "{fname}"] >= {max_value})'
            )
        else:
            field_cuts = (
                f'(obj["{ftype}", "{fname}"].in_units("{units}") <= {min_value}) | '
                f'(obj["{ftype}", "{fname}"].in_units("{units}") >= {max_value})'
            )
        cr = self.cut_region(field_cuts)
        return cr

    def include_inside(self, field, min_value, max_value, units=None):
        """
        This function will return a YTCutRegion where only the regions
        whose field are inside the interval from min_value to max_value are
        included.

        Parameters
        ----------
        field : string
            The field in which the conditional will be applied.
        min_value : float
            The minimum value inside the interval to be excluded.
        max_value : float
            The maximum value inside the interval to be excluded.
        units : string or None
            The units of the value threshold. None will use the default units
            given in the field.

        Returns
        -------
        cut_region : YTCutRegion
            The YTCutRegion with the field inside the given interval excluded.

        Examples
        --------
        >>> ds = yt.load("RedshiftOutput0005")
        >>> ad = ds.all_data()
        >>> cr = ad.include_inside(("gas", "temperature"), 1e5, 1e6)
        >>> print(cr.quantities.total_quantity(("gas", "cell_mass")).in_units("Msun"))
        """
        ftype, fname = self._determine_fields(field)[0]
        if units is None:
            field_cuts = (
                f'(obj["{ftype}", "{fname}"] > {min_value}) & '
                f'(obj["{ftype}", "{fname}"] < {max_value})'
            )
        else:
            field_cuts = (
                f'(obj["{ftype}", "{fname}"].in_units("{units}") > {min_value}) & '
                f'(obj["{ftype}", "{fname}"].in_units("{units}") < {max_value})'
            )
        cr = self.cut_region(field_cuts)
        return cr

    def exclude_outside(self, field, min_value, max_value, units=None):
        """
        This function will return a YTCutRegion where all of the regions
        whose field are outside the interval from min_value to max_value.

        Parameters
        ----------
        field : string
            The field in which the conditional will be applied.
        min_value : float
            The minimum value inside the interval to be excluded.
        max_value : float
            The maximum value inside the interval to be excluded.
        units : string or None
            The units of the value threshold. None will use the default units
            given in the field.

        Returns
        -------
        cut_region : YTCutRegion
            The YTCutRegion with the field outside the given interval excluded.

        Examples
        --------
        >>> ds = yt.load("RedshiftOutput0005")
        >>> ad = ds.all_data()
        >>> cr = ad.exclude_outside(("gas", "temperature"), 1e5, 1e6)
        >>> print(cr.quantities.total_quantity(("gas", "cell_mass")).in_units("Msun"))
        """
        cr = self.exclude_below(field, min_value, units)
        cr = cr.exclude_above(field, max_value, units)
        return cr

    def include_outside(self, field, min_value, max_value, units=None):
        """
        This function will return a YTCutRegion where only the regions
        whose field are outside the interval from min_value to max_value are
        included.

        Parameters
        ----------
        field : string
            The field in which the conditional will be applied.
        min_value : float
            The minimum value inside the interval to be excluded.
        max_value : float
            The maximum value inside the interval to be excluded.
        units : string or None
            The units of the value threshold. None will use the default units
            given in the field.

        Returns
        -------
        cut_region : YTCutRegion
            The YTCutRegion with the field outside the given interval excluded.

        Examples
        --------
        >>> ds = yt.load("RedshiftOutput0005")
        >>> ad = ds.all_data()
        >>> cr = ad.exclude_outside(("gas", "temperature"), 1e5, 1e6)
        >>> print(cr.quantities.total_quantity(("gas", "cell_mass")).in_units("Msun"))
        """
        cr = self.exclude_inside(field, min_value, max_value, units)
        return cr

    def exclude_below(self, field, value, units=None):
        """
        This function will return a YTCutRegion where all of the regions
        whose field is below a given value are masked.

        Parameters
        ----------
        field : string
            The field in which the conditional will be applied.
        value : float
            The minimum value that will not be masked in the output
            YTCutRegion.
        units : string or None
            The units of the value threshold. None will use the default units
            given in the field.

        Returns
        -------
        cut_region : YTCutRegion
            The YTCutRegion with the field below the given value masked.

        Examples
        --------
        >>> ds = yt.load("RedshiftOutput0005")
        >>> ad = ds.all_data()
        >>> cr = ad.exclude_below(("gas", "temperature"), 1e6)
        >>> print(cr.quantities.total_quantity(("gas", "cell_mass")).in_units("Msun"))
        """
        return self._build_operator_cut(">=", field, value, units)

    def exclude_nan(self, field, units=None):
        """
        This function will return a YTCutRegion where all of the regions
        whose field is NaN are masked.

        Parameters
        ----------
        field : string
            The field in which the conditional will be applied.
        units : string or None
            The units of the value threshold. None will use the default units
            given in the field.

        Returns
        -------
        cut_region : YTCutRegion
            The YTCutRegion with the NaN entries of the field masked.

        Examples
        --------
        >>> ds = yt.load("RedshiftOutput0005")
        >>> ad = ds.all_data()
        >>> cr = ad.exclude_nan(("gas", "temperature"))
        >>> print(cr.quantities.total_quantity(("gas", "cell_mass")).in_units("Msun"))
        """
        return self._build_function_cut("~np.isnan", field, units, locals={"np": np})

    def include_below(self, field, value, units=None):
        """
        This function will return a YTCutRegion where only the regions
        whose field is below a given value are included.

        Parameters
        ----------
        field : string
            The field in which the conditional will be applied.
        value : float
            The minimum value that will not be masked in the output
            YTCutRegion.
        units : string or None
            The units of the value threshold. None will use the default units
            given in the field.

        Returns
        -------
        cut_region : YTCutRegion
            The YTCutRegion with only regions with the field below the given
            value included.

        Examples
        --------
        >>> ds = yt.load("RedshiftOutput0005")
        >>> ad = ds.all_data()
        >>> cr = ad.include_below(("gas", "temperature"), 1e5, 1e6)
        >>> print(cr.quantities.total_quantity(("gas", "cell_mass")).in_units("Msun"))
        """
        return self._build_operator_cut("<", field, value, units)

    def extract_isocontours(
        self, field, value, filename=None, rescale=False, sample_values=None
    ):
        r"""This identifies isocontours on a cell-by-cell basis, with no
        consideration of global connectedness, and returns the vertices of the
        Triangles in that isocontour.

        This function simply returns the vertices of all the triangles
        calculated by the `marching cubes
        <https://en.wikipedia.org/wiki/Marching_cubes>`_ algorithm; for more
        complex operations, such as identifying connected sets of cells above a
        given threshold, see the extract_connected_sets function.  This is more
        useful for calculating, for instance, total isocontour area, or
        visualizing in an external program (such as `MeshLab
        <http://www.meshlab.net>`_.)

        Parameters
        ----------
        field : string
            Any field that can be obtained in a data object.  This is the field
            which will be isocontoured.
        value : float
            The value at which the isocontour should be calculated.
        filename : string, optional
            If supplied, this file will be filled with the vertices in .obj
            format.  Suitable for loading into meshlab.
        rescale : bool, optional
            If true, the vertices will be rescaled within their min/max.
        sample_values : string, optional
            Any field whose value should be extracted at the center of each
            triangle.

        Returns
        -------
        verts : array of floats
            The array of vertices, x,y,z.  Taken in threes, these are the
            triangle vertices.
        samples : array of floats
            If `sample_values` is specified, this will be returned and will
            contain the values of the field specified at the center of each
            triangle.

        Examples
        --------
        This will create a data object, find a nice value in the center, and
        output the vertices to "triangles.obj" after rescaling them.

        >>> dd = ds.all_data()
        >>> rho = dd.quantities["WeightedAverageQuantity"](
        ...     ("gas", "density"), weight=("gas", "cell_mass")
        ... )
        >>> verts = dd.extract_isocontours(
        ...     ("gas", "density"), rho, "triangles.obj", True
        ... )
        """
        from yt.data_objects.static_output import ParticleDataset
        from yt.frontends.stream.data_structures import StreamParticlesDataset

        verts = []
        samples = []
        if isinstance(self.ds, (ParticleDataset, StreamParticlesDataset)):
            raise NotImplementedError
        for block, mask in self.blocks:
            my_verts = self._extract_isocontours_from_grid(
                block, mask, field, value, sample_values
            )
            if sample_values is not None:
                my_verts, svals = my_verts
                samples.append(svals)
            verts.append(my_verts)
        verts = np.concatenate(verts).transpose()
        verts = self.comm.par_combine_object(verts, op="cat", datatype="array")
        verts = verts.transpose()
        if sample_values is not None:
            samples = np.concatenate(samples)
            samples = self.comm.par_combine_object(samples, op="cat", datatype="array")
        if rescale:
            mi = np.min(verts, axis=0)
            ma = np.max(verts, axis=0)
            verts = (verts - mi) / (ma - mi).max()
        if filename is not None and self.comm.rank == 0:
            if hasattr(filename, "write"):
                f = filename
            else:
                f = open(filename, "w")
            for v1 in verts:
                f.write(f"v {v1[0]:0.16e} {v1[1]:0.16e} {v1[2]:0.16e}\n")
            for i in range(len(verts) // 3):
                f.write(f"f {i * 3 + 1} {i * 3 + 2} {i * 3 + 3}\n")
            if not hasattr(filename, "write"):
                f.close()
        if sample_values is not None:
            return verts, samples
        return verts

    def _extract_isocontours_from_grid(
        self, grid, mask, field, value, sample_values=None
    ):
        vc_fields = [field]
        if sample_values is not None:
            vc_fields.append(sample_values)

        vc_data = grid.get_vertex_centered_data(vc_fields, no_ghost=False)
        try:
            svals = vc_data[sample_values]
        except KeyError:
            svals = None

        my_verts = march_cubes_grid(
            value, vc_data[field], mask, grid.LeftEdge, grid.dds, svals
        )
        return my_verts

    def calculate_isocontour_flux(
        self, field, value, field_x, field_y, field_z, fluxing_field=None
    ):
        r"""This identifies isocontours on a cell-by-cell basis, with no
        consideration of global connectedness, and calculates the flux over
        those contours.

        This function will conduct `marching cubes
        <https://en.wikipedia.org/wiki/Marching_cubes>`_ on all the cells in a
        given data container (grid-by-grid), and then for each identified
        triangular segment of an isocontour in a given cell, calculate the
        gradient (i.e., normal) in the isocontoured field, interpolate the local
        value of the "fluxing" field, the area of the triangle, and then return:

        area * local_flux_value * (n dot v)

        Where area, local_value, and the vector v are interpolated at the barycenter
        (weighted by the vertex values) of the triangle.  Note that this
        specifically allows for the field fluxing across the surface to be
        *different* from the field being contoured.  If the fluxing_field is
        not specified, it is assumed to be 1.0 everywhere, and the raw flux
        with no local-weighting is returned.

        Additionally, the returned flux is defined as flux *into* the surface,
        not flux *out of* the surface.

        Parameters
        ----------
        field : string
            Any field that can be obtained in a data object.  This is the field
            which will be isocontoured and used as the "local_value" in the
            flux equation.
        value : float
            The value at which the isocontour should be calculated.
        field_x : string
            The x-component field
        field_y : string
            The y-component field
        field_z : string
            The z-component field
        fluxing_field : string, optional
            The field whose passage over the surface is of interest.  If not
            specified, assumed to be 1.0 everywhere.

        Returns
        -------
        flux : float
            The summed flux.  Note that it is not currently scaled; this is
            simply the code-unit area times the fields.

        Examples
        --------
        This will create a data object, find a nice value in the center, and
        calculate the metal flux over it.

        >>> dd = ds.all_data()
        >>> rho = dd.quantities["WeightedAverageQuantity"](
        ...     ("gas", "density"), weight=("gas", "cell_mass")
        ... )
        >>> flux = dd.calculate_isocontour_flux(
        ...     ("gas", "density"),
        ...     rho,
        ...     ("gas", "velocity_x"),
        ...     ("gas", "velocity_y"),
        ...     ("gas", "velocity_z"),
        ...     ("gas", "metallicity"),
        ... )
        """
        flux = 0.0
        for block, mask in self.blocks:
            flux += self._calculate_flux_in_grid(
                block, mask, field, value, field_x, field_y, field_z, fluxing_field
            )
        flux = self.comm.mpi_allreduce(flux, op="sum")
        return flux

    def _calculate_flux_in_grid(
        self, grid, mask, field, value, field_x, field_y, field_z, fluxing_field=None
    ):
        vc_fields = [field, field_x, field_y, field_z]
        if fluxing_field is not None:
            vc_fields.append(fluxing_field)

        vc_data = grid.get_vertex_centered_data(vc_fields)

        if fluxing_field is None:
            ff = np.ones_like(vc_data[field], dtype="float64")
        else:
            ff = vc_data[fluxing_field]

        return march_cubes_grid_flux(
            value,
            vc_data[field],
            vc_data[field_x],
            vc_data[field_y],
            vc_data[field_z],
            ff,
            mask,
            grid.LeftEdge,
            grid.dds,
        )

    def extract_connected_sets(
        self, field, num_levels, min_val, max_val, log_space=True, cumulative=True
    ):
        """
        This function will create a set of contour objects, defined
        by having connected cell structures, which can then be
        studied and used to 'paint' their source grids, thus enabling
        them to be plotted.

        Note that this function *can* return a connected set object that has no
        member values.
        """
        if log_space:
            cons = np.logspace(np.log10(min_val), np.log10(max_val), num_levels + 1)
        else:
            cons = np.linspace(min_val, max_val, num_levels + 1)
        contours = {}
        for level in range(num_levels):
            contours[level] = {}
            if cumulative:
                mv = max_val
            else:
                mv = cons[level + 1]
            from yt.data_objects.level_sets.api import identify_contours
            from yt.data_objects.level_sets.clump_handling import add_contour_field

            nj, cids = identify_contours(self, field, cons[level], mv)
            unique_contours = set()
            for sl_list in cids.values():
                for _sl, ff in sl_list:
                    unique_contours.update(np.unique(ff))
            contour_key = uuid.uuid4().hex
            # In case we're a cut region already...
            base_object = getattr(self, "base_object", self)
            add_contour_field(base_object.ds, contour_key)
            for cid in sorted(unique_contours):
                if cid == -1:
                    continue
                contours[level][cid] = base_object.cut_region(
                    [f"obj['contours_{contour_key}'] == {cid}"],
                    {f"contour_slices_{contour_key}": cids},
                )
        return cons, contours

    def _get_bbox(self):
        """
        Return the bounding box for this data container.
        This generic version will return the bounds of the entire domain.
        """
        return self.ds.domain_left_edge, self.ds.domain_right_edge

    def get_bbox(self) -> tuple[unyt_array, unyt_array]:
        """
        Return the bounding box for this data container.
        """
        geometry: Geometry = self.ds.geometry
        if geometry is Geometry.CARTESIAN:
            le, re = self._get_bbox()
            return (le.to("code_length"), re.to("code_length"))
        elif (
            geometry is Geometry.CYLINDRICAL
            or geometry is Geometry.POLAR
            or geometry is Geometry.SPHERICAL
            or geometry is Geometry.GEOGRAPHIC
            or geometry is Geometry.INTERNAL_GEOGRAPHIC
            or geometry is Geometry.SPECTRAL_CUBE
        ):
            raise NotImplementedError(
                f"get_bbox is currently not implemented for {geometry=}!"
            )
        else:
            assert_never(geometry)

    def volume(self):
        """
        Return the volume of the data container.
        This is found by adding up the volume of the cells with centers
        in the container, rather than using the geometric shape of
        the container, so this may vary very slightly
        from what might be expected from the geometric volume.
        """
        return self.quantities.total_quantity(("index", "cell_volume"))



import inspect

import numpy as np

import yt.frontends as frontends_module
from yt.config import ytcfg
from yt.fields.derived_field import NullFunc
from yt.frontends.api import _frontends
from yt.frontends.stream.fields import StreamFieldInfo
from yt.funcs import obj_length
from yt.testing import fake_random_ds
from yt.units import dimensions
from yt.units.yt_array import Unit
from yt.utilities.cosmology import Cosmology

fields, units = [], []

for fname, (code_units, _aliases, _dn) in StreamFieldInfo.known_other_fields:
    fields.append(("gas", fname))
    units.append(code_units)
base_ds = fake_random_ds(4, fields=fields, units=units)
base_ds.index
base_ds.cosmological_simulation = 1
base_ds.cosmology = Cosmology()


ytcfg["yt", "internals", "within_testing"] = True
np.seterr(all="ignore")


def _strip_ftype(field):
    if not isinstance(field, tuple):
        return field
    elif field[0] == "all":
        return field
    return field[1]


np.random.seed(int(0x4D3D3D3))
units = [base_ds._get_field_info(f).units for f in fields]
fields = [_strip_ftype(f) for f in fields]
ds = fake_random_ds(16, fields=fields, units=units, particles=1)
ds.parameters["HydroMethod"] = "streaming"
ds.parameters["EOSType"] = 1.0
ds.parameters["EOSSoundSpeed"] = 1.0
ds.conversion_factors["Time"] = 1.0
ds.conversion_factors.update({f: 1.0 for f in fields})
ds.gamma = 5.0 / 3.0
ds.current_redshift = 0.0001
ds.cosmological_simulation = 1
ds.hubble_constant = 0.7
ds.omega_matter = 0.27
ds.omega_lambda = 0.73
ds.cosmology = Cosmology(
    hubble_constant=ds.hubble_constant,
    omega_matter=ds.omega_matter,
    omega_lambda=ds.omega_lambda,
    unit_registry=ds.unit_registry,
)
for my_unit in ["m", "pc", "AU", "au"]:
    new_unit = f"{my_unit}cm"
    my_u = Unit(my_unit, registry=ds.unit_registry)
    ds.unit_registry.add(
        new_unit,
        my_u.base_value,
        dimensions.length,
        "\\rm{%s}/(1+z)" % my_unit,
        prefixable=True,
    )


header = r"""
.. _field-list:

Field List
==========

This is a list of many of the fields available in yt.  We have attempted to
include most of the fields that are accessible through the plugin system, as
well as the fields that are known by the frontends, however it is possible to
generate many more permutations, particularly through vector operations. For
more information about the fields framework, see :ref:`fields`.

Some fields are recognized by specific frontends only. These are typically
fields like density and temperature that have their own names and units in
the different frontend datasets. Often, these fields are aliased to their
yt-named counterpart fields (typically 'gas' fieldtypes). For example, in
the ``FLASH`` frontend, the ``dens`` field (i.e. ``(flash, dens)``) is aliased
to the gas field density (i.e. ``(gas, density)``), similarly ``(flash, velx)``
is aliased to ``(gas, velocity_x)``, and so on. In what follows, if a field
is aliased it will be noted.

Try using the ``ds.field_list`` and ``ds.derived_field_list`` to view the
native and derived fields available for your dataset respectively. For example
to display the native fields in alphabetical order:

.. notebook-cell::

  import yt
  ds = yt.load("Enzo_64/DD0043/data0043")
  for i in sorted(ds.field_list):
    print(i)

To figure out out what all of the field types here mean, see
:ref:`known-field-types`.

.. contents:: Table of Contents
   :depth: 1
   :local:
   :backlinks: none

.. _yt-fields:

Universal Fields
----------------
"""

footer = """

Index of Fields
---------------

.. contents::
   :depth: 3
   :backlinks: none

"""
print(header)

seen = []


def fix_units(units, in_cgs=False):
    unit_object = Unit(units, registry=ds.unit_registry)
    if in_cgs:
        unit_object = unit_object.get_cgs_equivalent()
    latex = unit_object.latex_representation()
    return latex.replace(r"\ ", "~")


def print_all_fields(fl):
    for fn in sorted(fl):
        df = fl[fn]
        f = df._function
        s = f"{df.name}"
        print(s)
        print("^" * len(s))
        print()
        if obj_length(df.units) > 0:
            # Most universal fields are in CGS except for these special fields
            if df.name[1] in [
                "particle_position",
                "particle_position_x",
                "particle_position_y",
                "particle_position_z",
                "entropy",
                "kT",
                "metallicity",
                "dx",
                "dy",
                "dz",
                "cell_volume",
                "x",
                "y",
                "z",
            ]:
                print(f"   * Units: :math:`{fix_units(df.units)}`")
            else:
                print(f"   * Units: :math:`{fix_units(df.units, in_cgs=True)}`")
        print(f"   * Sampling Method: {df.sampling_type}")
        print()
        print("**Field Source**")
        print()
        if f == NullFunc:
            print("No source available.")
            print()
            continue
        else:
            print(".. code-block:: python")
            print()
            for line in inspect.getsource(f).split("\n"):
                print("  " + line)
            print()


ds.index
print_all_fields(ds.field_info)


class FieldInfo:
    """a simple container to hold the information about fields"""

    def __init__(self, ftype, field, ptype):
        name = field[0]
        self.units = ""
        u = field[1][0]
        if len(u) > 0:
            self.units = r":math:`\mathrm{%s}`" % fix_units(u)
        a = [f"``{f}``" for f in field[1][1] if f]
        self.aliases = " ".join(a)
        self.dname = ""
        if field[1][2] is not None:
            self.dname = f":math:`{field[1][2]}`"

        if ftype != "particle_type":
            ftype = f"'{ftype}'"
        self.name = f"({ftype}, '{name}')"
        self.ptype = ptype


current_frontends = [f for f in _frontends if f not in ["stream"]]

for frontend in current_frontends:
    this_f = getattr(frontends_module, frontend)
    field_info_names = [fi for fi in dir(this_f) if "FieldInfo" in fi]
    dataset_names = [dset for dset in dir(this_f) if "Dataset" in dset]

    if frontend == "gadget":
        # Drop duplicate entry for GadgetHDF5, add special case for FieldInfo
        # entry
        dataset_names = ["GadgetDataset"]
        field_info_names = ["GadgetFieldInfo"]
    elif frontend == "boxlib":
        field_info_names = []
        for d in dataset_names:
            if "Maestro" in d:
                field_info_names.append("MaestroFieldInfo")
            elif "Castro" in d:
                field_info_names.append("CastroFieldInfo")
            else:
                field_info_names.append("BoxlibFieldInfo")
    elif frontend == "chombo":
        # remove low dimensional field info containers for ChomboPIC
        field_info_names = [
            f for f in field_info_names if "1D" not in f and "2D" not in f
        ]

    for dset_name, fi_name in zip(dataset_names, field_info_names):
        fi = getattr(this_f, fi_name)
        nfields = 0
        if hasattr(fi, "known_other_fields"):
            known_other_fields = fi.known_other_fields
            nfields += len(known_other_fields)
        else:
            known_other_fields = []
        if hasattr(fi, "known_particle_fields"):
            known_particle_fields = fi.known_particle_fields
            if "Tipsy" in fi_name:
                known_particle_fields += tuple(fi.aux_particle_fields.values())
            nfields += len(known_particle_fields)
        else:
            known_particle_fields = []
        if nfields > 0:
            print(f".. _{dset_name.replace('Dataset', '')}_specific_fields:\n")
            h = f"{dset_name.replace('Dataset', '')}-Specific Fields"
            print(h)
            print("-" * len(h) + "\n")

            field_stuff = []
            for field in known_other_fields:
                field_stuff.append(FieldInfo(frontend, field, False))
            for field in known_particle_fields:
                if frontend in ["sph", "halo_catalogs", "sdf"]:
                    field_stuff.append(FieldInfo("particle_type", field, True))
                else:
                    field_stuff.append(FieldInfo("io", field, True))

            # output
            len_name = 10
            len_units = 5
            len_aliases = 7
            len_part = 9
            len_disp = 12
            for f in field_stuff:
                len_name = max(len_name, len(f.name))
                len_aliases = max(len_aliases, len(f.aliases))
                len_units = max(len_units, len(f.units))
                len_disp = max(len_disp, len(f.dname))

            fstr = "{nm:{nw}}  {un:{uw}}  {al:{aw}}  {pt:{pw}}  {dp:{dw}}"
            header = fstr.format(
                nm="field name",
                nw=len_name,
                un="units",
                uw=len_units,
                al="aliases",
                aw=len_aliases,
                pt="particle?",
                pw=len_part,
                dp="display name",
                dw=len_disp,
            )

            div = fstr.format(
                nm="=" * len_name,
                nw=len_name,
                un="=" * len_units,
                uw=len_units,
                al="=" * len_aliases,
                aw=len_aliases,
                pt="=" * len_part,
                pw=len_part,
                dp="=" * len_disp,
                dw=len_disp,
            )
            print(div)
            print(header)
            print(div)

            for f in field_stuff:
                print(
                    fstr.format(
                        nm=f.name,
                        nw=len_name,
                        un=f.units,
                        uw=len_units,
                        al=f.aliases,
                        aw=len_aliases,
                        pt=f.ptype,
                        pw=len_part,
                        dp=f.dname,
                        dw=len_disp,
                    )
                )

            print(div)
            print("")

print(footer)
